#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
ALBACalibration

[Description]
HwObj used to grab the zoom/pixel size calibration from
PySignal simulator (TangoDS).


Example Hardware Object XML file :
==================================
<object class="ALBACalibration">
  <username>Calibration</username>
  <taurusname>bl13/ct/variables</taurusname>
  <channel type="sardana" name="calibx">OAV_PIXELSIZE_X</channel>
  <channel type="sardana" name="caliby">OAV_PIXELSIZE_Y</channel>
  <interval>200</interval>
  <threshold>0.001</threshold>
</object>
"""

import logging

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR

__author__ = "Jordi Andreu"
__credits__ = ["MXCuBE collaboration"]

__version__ = "2.2."
__maintainer__ = "Jordi Andreu"
__email__ = "jandreu[at]cells.es"
__status__ = "Draft"


class ALBACalibration(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.calibx = self.get_channel_object("calibx")
        self.caliby = self.get_channel_object("caliby")

        if self.calibx is not None and self.caliby is not None:
            logging.getLogger().info("Connected to pixel size calibration channels")

    def getCalibration(self):
        calibx = self.calibx.get_value()
        caliby = self.caliby.get_value()
        logging.getLogger().debug(
            "Returning calibration: x=%s, y=%s" % (calibx, caliby)
        )
        return [calibx, caliby]


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    calib = hwr.get_hardware_object("/calibration")
    print("Calibration is: ", calib.getCalibration())


if __name__ == "__main__":
    test()



"""ESRF SC3 Sample Changer Hardware Object"""

import functools

import SC3

from mxcubecore.TaskUtils import (
    cleanup,
    error_cleanup,
    task,
)


class ESRFSC3(SC3.SC3):
    (FLAG_SC_IN_USE, FLAG_MINIDIFF_CAN_MOVE, FLAG_SC_CAN_LOAD, FLAG_SC_NEVER) = (
        1,
        2,
        4,
        8,
    )
    (STATE_BASKET_NOTPRESENT, STATE_BASKET_UNKNOWN, STATE_BASKET_PRESENT) = (-1, 0, 1)

    def __init__(self, *args, **kwargs):
        SC3.SC3.__init__(self, *args, **kwargs)

        self.lastOperationalFlags = ESRFSC3.FLAG_SC_NEVER

    def init(self):
        SC3.SC3.init(self)

        try:
            operationalChan = self.get_channel_object("OperationalFlags")
            chan = operationalChan.get_value()
            operationalChan.connect_signal("update", self.operationalFlagsChanged)
        except Exception:
            operationalChan = self.get_property("OperationalFlags")
            chan = operationalChan

        try:
            self.operationalFlagsChanged(chan)
        except Exception:
            self.log.exception(
                "%s: error getting SC vs MD operational flags" % self.name()
            )

        try:
            self.microdiff
        except AttributeError:
            self.prepareCentringAfterLoad = True
        else:
            self.prepareCentringAfterLoad = False

    def __getSample(self, sample_id, sample_location):
        if sample_id and sample_location:
            self.log.debug(
                "%s: both sample barcode and location provided, discarding barcode...",
                self.name(),
            )
            sample_id = None

        if sample_id:
            sample = self.get_component_by_id(sample_id)
        else:
            if sample_location:
                basket_number, sample_number = sample_location
                sample = self.get_component_by_address(
                    SC3.Pin.get_sample_address(basket_number, sample_number)
                )
            else:
                sample = self.get_selected_sample()

        return sample

    def chained_load(self, sample_to_unload, sample_to_load):
        SC3.SC3.unload(self, sample_to_unload)
        return SC3.SC3.load(self, sample_to_load)

    @task
    def load(
        self,
        holderLength,
        sample_id=None,
        sample_location=None,
        sampleIsLoadedCallback=None,
        failureCallback=None,
        prepareCentring=True,
    ):
        loaded = False

        with error_cleanup(
            functools.partial(self.emit, "stateChanged", SC3.SampleChangerState.Alarm),
            failureCallback,
        ):
            with cleanup(
                functools.partial(
                    self.emit, "stateChanged", SC3.SampleChangerState.Ready
                )
            ):
                with cleanup(self.unlockMinidiffMotors, wait=True, timeout=3):
                    loaded = self.__load_sample(
                        holderLength, sample_id, sample_location
                    )

                if loaded:
                    self.log.debug("%s: sample is loaded", self.name())

                    if self.prepareCentringAfterLoad and prepareCentring:
                        self.log.debug(
                            "%s: preparing minidiff for sample centring", self.name()
                        )
                        self.emit("stateChanged", SC3.SampleChangerState.Moving)
                        self.emit(
                            "statusChanged", "Preparing minidiff for sample centring"
                        )
                        self.prepare_centring(wait=True, timeout=1000)

                    self.emit("statusChanged", "Ready")

                    if callable(sampleIsLoadedCallback):
                        sampleIsLoadedCallback()

    def _getLoadingState(self):
        # if needed, should wait for SC to be able to load (loading state)
        pass

    def __load_sample(self, holderLength, sample_id, sample_location):
        self.log.debug("%s: in load_sample", self.name())

        sample = self.__getSample(sample_id, sample_location)

        if self.get_loaded_sample() == sample:
            return True

        if not holderLength:
            holderLength = 22
            self.log.debug(
                "%s: loading sample: using default holder length (%d mm)",
                self.name(),
                holderLength,
            )

        sample._set_holder_length(holderLength)

        self.emit("stateChanged", SC3.SampleChangerState.Moving)
        self.emit("statusChanged", "Moving diffractometer to loading position")
        self._moveToLoadingPosition(holderLength, wait=True, timeout=10000)

        try:
            SC3.SC3.load(self, sample, wait=True)
        except Exception as err:
            self.emit("statusChanged", str(err))
            raise

        self._getLoadingState()
        return self.get_loaded_sample() == sample

    def load_sample(self, *args, **kwargs):
        kwargs["wait"] = True
        return self.load(*args, **kwargs)

    @task
    def unload(
        self,
        holderLength,
        sample_id=None,
        sample_location=None,
        sampleIsUnloadedCallback=None,
        failureCallback=None,
    ):
        unloaded = False

        with error_cleanup(
            functools.partial(self.emit, "stateChanged", SC3.SampleChangerState.Alarm),
            failureCallback,
        ):
            with cleanup(
                functools.partial(
                    self.emit, "stateChanged", SC3.SampleChangerState.Ready
                )
            ):
                with cleanup(self.unlockMinidiffMotors, wait=True, timeout=3):
                    unloaded = self.__unload_sample(
                        holderLength, sample_id, sample_location
                    )

                if unloaded:
                    self.log.debug("%s: sample has been unloaded", self.name())

                    self.emit("statusChanged", "Ready")

                    if callable(sampleIsUnloadedCallback):
                        sampleIsUnloadedCallback()

    def __unload_sample(self, holderLength, sample_id, sample_location):
        sample = self.__getSample(sample_id, sample_location)

        if not holderLength:
            holderLength = 22
            self.log.debug(
                "%s: unloading sample: using default holder length (%d mm)",
                self.name(),
                holderLength,
            )

        sample._set_holder_length(holderLength)

        self.emit("stateChanged", SC3.SampleChangerState.Moving)
        self.emit("statusChanged", "Moving diffractometer to unloading position")

        self._moveToUnloadingPosition(holderLength, wait=True, timeout=10000)
        SC3.SC3.unload(self, sample, wait=True)

        self._getLoadingState()
        return not self.has_loaded_sample()

    def moveCryoIn(self):
        cryoDevice = self.get_deviceby_role("Cryo")
        if cryoDevice is not None:
            cryoDevice.wagoIn()

    def moveLightIn(self):
        lightDevice = self.get_deviceby_role("Light")
        if lightDevice is not None:
            lightDevice.wagoIn()

    # def getSCHolderLength(self):
    #    return

    def is_microdiff(self):
        return not self.prepareCentringAfterLoad

    def operationalFlagsChanged(self, val):
        try:
            val = int(val)
        except Exception:
            self.log.exception("%s: error reading operational flags" % self.name())
            return

        old_sc_can_load = self.lastOperationalFlags & ESRFSC3.FLAG_SC_CAN_LOAD
        new_sc_can_load = val & ESRFSC3.FLAG_SC_CAN_LOAD
        if old_sc_can_load != new_sc_can_load:
            self.emit("sampleChangerCanLoad", (new_sc_can_load > 0,))

        old_md_can_move = self.lastOperationalFlags & ESRFSC3.FLAG_MINIDIFF_CAN_MOVE
        new_md_can_move = val & ESRFSC3.FLAG_MINIDIFF_CAN_MOVE
        if old_md_can_move != new_md_can_move:
            self.emit("minidiffCanMove", (new_md_can_move > 0,))

        old_sc_in_use = self.lastOperationalFlags & ESRFSC3.FLAG_SC_IN_USE
        new_sc_in_use = val & ESRFSC3.FLAG_SC_IN_USE
        if old_sc_in_use != new_sc_in_use:
            self.emit("sampleChangerInUse", (new_sc_in_use > 0,))

        self.lastOperationalFlags = val

    def sampleChangerInUse(self):
        sc_in_use = self.lastOperationalFlags & ESRFSC3.FLAG_SC_IN_USE
        return sc_in_use > 0

    def sampleChangerCanLoad(self):
        sc_can_load = self.lastOperationalFlags & ESRFSC3.FLAG_SC_CAN_LOAD
        return sc_can_load > 0

    def minidiffCanMove(self):
        md_can_move = self.lastOperationalFlags & ESRFSC3.FLAG_MINIDIFF_CAN_MOVE
        return md_can_move > 0

    def minidiffGetControl(self):
        try:
            self.unlockMinidiffMotors()
        except Exception:
            self.log.exception("%s: error unlocking minidiff motors" % self.name())
            return False
        return True

    """
    def sampleChangerToLoadingPosition(self):
        try:
            r=self._moveToLoadingPosition()
        except:
            self.log.exception("%s: error moving sample changer to loading position" % self.name())
            return False
        return True
    """



from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractMCA import AbstractMCA
from mxcubecore.TaskUtils import task


class BlissRontecMCA(AbstractMCA, HardwareObject):
    def __init__(self, name):
        AbstractMCA.__init__(self)
        HardwareObject.__init__(self, name)
        self.mca = None
        self.calib_cf = []

    def init(self):
        session = self.get_object_by_role("bliss_session")
        obj_name = self.get_property("object_name")
        self.mca = getattr(session, obj_name)

    @task
    def read_raw_data(self, chmin=0, chmax=4095, save_data=False):
        return self.mca.read_raw_data(chmin, chmax, save_data)

    @task
    def read_roi_data(self, save_data=False):
        return self.mca.read_roi_data(save_data)

    @task
    def read_data(self, chmin=0, chmax=4095, calib=False, save_data=False):
        return self.mca.read_data(chmin, chmax, calib, save_data)

    @task
    def set_calibration(self, fname=None, calib_cf=None):
        calib = None
        if fname:
            calib = fname
        elif calib_cf:
            calib = calib_cf
        return self.mca.set_calibration(calib)

    @task
    def get_calibration(self):
        return self.mca.get_calibration()

    @task
    def set_roi(self, emin, emax, **kwargs):
        self.mca.set_roi(emin, emax, **kwargs)

    @task
    def get_roi(self, **kwargs):
        return self.mca.get_roi(**kwargs)

    @task
    def clear_roi(self, **kwargs):
        self.mca.clear_roi(**kwargs)

    @task
    def get_times(self):
        return self.mca.get_times()

    @task
    def get_presets(self, **kwargs):
        return self.mca.get_presets(**kwargs)

    @task
    def set_presets(self, **kwargs):
        self.mca.set_presets(**kwargs)

    @task
    def start_acq(self, cnt_time=None):
        self.mca.start_acq(cnt_time)

    @task
    def stop_acq(self):
        self.mca.stop_acq()

    @task
    def clear_spectrum(self):
        self.mca.clear_spectrum()



import time

from gevent.event import AsyncResult

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.GenericDiffractometer import GenericDiffractometer


class LNLSDiffractometer(GenericDiffractometer):
    def __init__(self, name):
        GenericDiffractometer.__init__(self, name)

    def init(self):
        GenericDiffractometer.init(self)
        self._bluesky_api = HWR.beamline.get_object_by_role("bluesky")
        self.pixels_per_mm_x = 10**-4
        self.pixels_per_mm_y = 10**-4
        self.beam_position = [318, 238]
        self.last_centred_position = self.beam_position
        self.current_motor_positions = {
            "phiy": 0,
            "sampx": 0,
            "sampy": 0,
            "zoom": 0,
            "focus": 0,
            "phiz": 0,
            "phi": 0,
            "kappa": 0,
            "kappa_phi": 0,
        }

        self.centring_time = 0
        self.mount_mode = self.get_property("sample_mount_mode")
        if self.mount_mode is None:
            self.mount_mode = "manual"

        self.connect(self.motor_hwobj_dict["phi"], "valueChanged", self.phi_motor_moved)
        self.connect(
            self.motor_hwobj_dict["phiy"], "valueChanged", self.phiy_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["phiz"], "valueChanged", self.phiz_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa"], "valueChanged", self.kappa_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa_phi"],
            "valueChanged",
            self.kappa_phi_motor_moved,
        )
        self.connect(
            self.motor_hwobj_dict["sampx"], "valueChanged", self.sampx_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["sampy"], "valueChanged", self.sampy_motor_moved
        )

    def is_ready(self) -> bool:
        return True

    def phi_motor_moved(self, pos):
        self.current_motor_positions["phi"] = pos
        self.emit("phiMotorMoved", pos)

    def phiy_motor_moved(self, pos):
        self.current_motor_positions["phiy"] = pos

    def phiz_motor_moved(self, pos):
        self.current_motor_positions["phiz"] = pos

    def sampx_motor_moved(self, pos):
        self.current_motor_positions["sampx"] = pos

    def sampy_motor_moved(self, pos):
        self.current_motor_positions["sampy"] = pos

    def kappa_motor_moved(self, pos):
        self.current_motor_positions["kappa"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaMotorMoved", pos)

    def kappa_phi_motor_moved(self, pos):
        self.current_motor_positions["kappa_phi"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaPhiMotorMoved", pos)

    def manual_centring(self):
        self.log.info("Initializing manual sample alignment...")
        for step in range(3):
            self.log.info(f"Step {step + 1} of 3...")
            self.user_clicked_event = AsyncResult()
            self.waiting_for_click = True
            x, y = self.user_clicked_event.get()
            self.log.info(f"{x}, {y}")
            self._bluesky_api.execute_plan(
                plan_name="manual_alignment",
                kwargs={"x_px": x, "y_px": y, "step": step},
            )
        self.log.info("Manual sample alignment has finished...")
        return {}

    def automatic_centring(self):
        self.log.info("Initializing automatic sample alignment...")
        self._bluesky_api.execute_plan(plan_name="automatic_alignment")
        self.log.info("Automatic sample alignment has finished...")

    def move_to_beam(self, x, y, omega=None):
        self.log.info("Moving to beam...")

        self._bluesky_api.execute_plan(
            plan_name="move_to_beam",
            kwargs={
                "x_px": x - self.beam_position[0],
                "y_px": y - self.beam_position[1],
            },
        )
        self.log.info("Move to beam has finished...")

    def motor_positions_to_screen(self, motor_positions):
        return self.beam_position

    def get_value_motors(self):
        return self.current_motor_positions



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
BeamDefiner ESRF implementation class - methods to define the size and shape of
the beam.
Example yml configuration:

.. code-block:: yaml

 class: ESRF.ESRFBeam.ESRFBeam
 configuration:
   beam_divergence_horizontal: 104
   beam_divergence_vertical: 6.5
   definer_type: definer   # could be also aperture or slits
 objects:
   aperture: udiff_aperture.yml
   definer: beam_definer.yml
   monitor_beam: monitor_beam.yaml
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractBeam import AbstractBeam, BeamShape


class ESRFBeam(AbstractBeam):
    """Beam ESRF implementation"""

    unit = "mm"

    def __init__(self, name) -> None:
        super().__init__(name)
        self._beam_check_obj = None
        self._monitorbeam_obj = None

    def init(self) -> None:
        """Initialize hardware"""
        super().init()

        _definer_type = []
        self._aperture = self.get_object_by_role("aperture")
        if self._aperture:
            _definer_type.append("aperture")

        _slits = self.get_property("slits")
        _bliss_obj = self.get_object_by_role("bliss")
        if _slits:
            self._slits = {}
            _definer_type.append("slits")
            for name in _slits.split():
                _key, _val = name.split(":")
                self._slits.update({_key: _bliss_obj.getattribute(_val)})

        self._definer = self.get_object_by_role("definer")
        if self._definer:
            _definer_type.append("definer")

        if len(_definer_type) == 1:
            self._definer_type = _definer_type[0]
        else:
            self._definer_type = None

        self._definer_type = self.get_property("definer_type") or self._definer_type

        beam_position = self.get_property("beam_position")

        if beam_position:
            self._beam_position_on_screen = tuple(map(float, beam_position.split()))

        if self._aperture:
            self._aperture.connect("valueChanged", self._re_emit_values)
            self._aperture.connect("stateChanged", self._re_emit_values)

        if self._definer:
            self._definer.connect("valueChanged", self._re_emit_values)
            self._definer.connect("stateChanged", self._re_emit_values)

        self._monitorbeam_obj = self.get_object_by_role("monitor_beam")

        if self._monitorbeam_obj:
            beam_check = self._monitorbeam_obj.actuator_name
            if beam_check and _bliss_obj:
                self._beam_check_obj = getattr(_bliss_obj, beam_check)

    def _re_emit_values(self, value):
        """Redefine as re_emit_values takes no arguments"""
        self.re_emit_values()

    def _get_aperture_value(self) -> tuple[list[float, float], str]:
        """Get the size and the label of the aperture in place.

        Returns:
            Size [mm] [width, height], label.
        """
        _size = self.aperture.get_value().value[1]
        try:
            _label = self.aperture.get_value().name
        except AttributeError:
            _label = str(_size)
        _size /= 1000.0

        return [_size, _size], _label

    def _get_definer_value(self) -> tuple[list[float, float], str]:
        """Get the size and the name of the definer in place.

        Returns:
            Size [mm] [width, height], label.
        """
        try:
            value = self.definer.get_value()
            if isinstance(value, tuple):
                return [value[1], value[1]], value[0]
            if value.name != "UNKNOWN":
                return list(value.value), value.name
        except AttributeError:
            self.log.info("Could not read beam size")
        return [-1, -1], "UNKNOWN"

    def _get_slits_size(self) -> dict[str, float]:
        """Get the size of the slits in place.

        Returns:
            ``{"width": float, "height": float}``.
        """
        beam_size = {}
        for _key, _val in self.slits:
            beam_size.update({_key: abs(_val.position)})
        return beam_size

    def _get_value(self) -> tuple[float, float, BeamShape, str]:
        """Get the size (width and height) of the beam, its shape and
           its label. The size is in millimeters.

        Returns:
            Four-item tuple: width, height, shape, name.
        """
        labels = {}
        _label = "UNKNOWN"
        if self.aperture:
            _size, _name = self._get_aperture_value()
            self._beam_size_dict.update({"aperture": _size})
            labels.update({"aperture": _name})

        if self.slits:
            _size, _name = self._get_slits_value()
            self._beam_size_dict.update({"slits": _size})
            labels.update({"slits": _name})

        if self.definer:
            _size, _name = self._get_definer_value()
            self._beam_size_dict.update({"definer": _size})
            labels.update({"definer": _name})

        info_dict = self.evaluate_beam_info()

        try:
            _label = labels[info_dict["label"]]
            self._beam_info_dict["label"] = _label
        except KeyError:
            _label = info_dict["label"]

        return self._beam_width, self._beam_height, self._beam_shape, _label

    def get_value_xml(self) -> tuple[float, float, str, str]:
        """XMLRPC does not handle Enum, the shape is transformed to string.

        Returns:
            Four-item tuple: width, height, shape, name
        """
        beamsize = self.get_value()
        return beamsize[0], beamsize[1], beamsize[2].value, beamsize[3]

    def get_available_size(self) -> dict:
        """Get the available predefined beam definer configuration.
        Returns:
            ``{"type": ["aperture"], "values": [labels]}`` or
            ``{"type": ["definer"], "values": [labels]}`` or
            ``{"type": ["width", "height"], "values":
                     [low_lim_w, high_lim_w, low_lim_h, high_lim_h]}``
        """
        if self._definer_type == "aperture":
            return {
                "type": ["aperture"],
                "values": self.aperture.get_diameter_size_list(),
            }

        if self._definer_type == "definer":
            return {
                "type": ["definer"],
                "values": self.definer.get_predefined_positions_list(),
            }

        if self._definer_type in (self.slits, "slits"):
            # get the list of the slits motors range
            _low_w, _high_w = self.slits["width"].get_limits()
            _low_h, _high_h = self.slits["height"].get_limits()
            return {
                "type": ["width", "height"],
                "values": [_low_w, _high_w, _low_h, _high_h],
            }

        return {}

    def get_defined_beam_size(self) -> dict:
        """Get the predefined beam labels and size.

        Returns:
            Dictionary with list of available beam size labels and
            the corresponding size (width,height) tuples.
            ``{"label": [str, str, ...], "size": [(w,h), (w,h), ...]}``
        """
        labels = []
        values = []

        if self._definer_type == "slits":
            # get the list of the slits motors range
            _low_w, _high_w = self.slits["width"].get_limits()
            _low_h, _high_h = self.slits["height"].get_limits()
            return {
                "label": ["low", "high"],
                "size": [(_low_w, _low_h), (_high_w, _high_h)],
            }

        if self._definer_type == "aperture":
            _enum = self.aperture.VALUES
        elif self._definer_type == "definer":
            _enum = self.definer.VALUES
        elif "aperture" in self._definer_type:
            _enum = self.aperture.VALUES

        for value in _enum:
            _nam = value.name
            if _nam not in ["IN", "OUT", "UNKNOWN"]:
                labels.append(_nam)
                if isinstance(value.value, tuple):
                    values.append(value.value)
                else:
                    values.append(value.value[0])
        return {"label": labels, "size": values}

    def _set_slits_size(self, size: list[float, float]):
        """Move the slits to the desired position.

        Args:
            Two-items list:  width, height in millimeters.

        Raises:
            RuntimeError: Size out of the limits.
               TypeError: Invalid size
        """

        if not isinstance(size, list):
            msg = "Incorrect input value for slits"
            raise TypeError(msg)
        w_lim = self.slits["width"].get_limits()
        h_lim = self.slits["height"].get_limits()
        try:
            msg = "Size out of the limits"
            if min(w_lim) > size[0] > max(w_lim):
                raise RuntimeError(msg)
            if min(h_lim) > size[1] > max(h_lim):
                raise RuntimeError(msg)
            self.slits["width"].set_value(size[0])
            self.slits["height"].set_value(size[1])
        except TypeError as err:
            raise TypeError("Invalid size") from err

    def _set_aperture_size(self, size: str):
        """Move the aperture to the desired size.

        Args:
            The position name.

        Raises:
            TypeError
        """
        if not isinstance(size, str):
            msg = "Incorrect input value for aperture"
            raise TypeError(msg)

        try:
            _ap = self.aperture.VALUES[size]
        except KeyError:
            _ap = self.aperture.VALUES[f"A{size}"]

        self.aperture.set_value(_ap)

    def _set_definer_size(self, size: str):
        """Move the definer to the desired size.

        Args:
            The position name.

        Raises:
            TypeError: Invalid size.
        """
        if not isinstance(size, str):
            msg = "Incorrect input value for definer"
            raise TypeError(msg)

        self._definer.set_value(self.definer.VALUES[size])

    def set_value(self, size: list[float] | str | None = None):
        """Set the beam size.

        Args:
            Two-items list: width and height for slits or
            aperture or definer name.

        Raises:
            RuntimeError: Beam definer not configured
                          Size out of the limits.
        """
        if self._definer_type in (self.slits, "slits"):
            self._set_slits_size(size)

        if self._definer_type in (self.aperture, "aperture"):
            self._set_aperture_size(size)

        if self._definer_type in (self.definer, "definer"):
            self._set_definer_size(size)

    def get_beam_position_on_screen(self) -> list[int, int]:
        """Get the beam position.

        Returns:
            X and Y coordinates of the beam position in pixels.
        """
        try:
            _beam_position_on_screen = HWR.beamline.diffractometer.get_beam_position()
        except AttributeError:
            msg = "Could not read beam position from MD, using OAV center"
            self.log.warning(msg)
            _beam_position_on_screen = (
                HWR.beamline.sample_view.camera.get_width() / 2,
                HWR.beamline.sample_view.camera.get_height() / 2,
            )

        self._beam_position_on_screen = _beam_position_on_screen

        return self._beam_position_on_screen

    def get_beam_size(self) -> tuple[float, float]:
        """Get the beam size.

        Returns:
            Two-item tuple: width and height.
        """
        beam_value = self.get_value()
        return beam_value[0], beam_value[1]

    def _is_beam(self) -> bool:
        """Check if there is beam.

        Returns:
            ``True`` if beam present, ``False`` otherwise
        """
        if self._beam_check_obj:
            return self._beam_check_obj.is_beam()
        return True

    def wait_for_beam(self, timeout: float | None = None):
        """Wait until beam present.

        Args:
            Optional - timeout in seconds,
            If timeout == 0: return at once and do not wait.
            if timeout is None: wait forever (default).

        Raises:
            RuntileError if no beam after the specified timeout.
        """
        if self._monitorbeam_obj:
            try:
                timeout = timeout or self._beam_check_obj.timeout
                if self._monitorbeam_obj.get_value().value:
                    self._beam_check_obj.wait_for_beam(timeout)
            except AttributeError:
                self.log.exception("")



"""
CATS maintenance commands hardware object.

Functionality in addition to sample-transfer functionality: power control,
lid control, error-recovery commands, ...
"""

import logging
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task

__author__ = "Michael Hellmig"
__credits__ = ["The MxCuBE collaboration"]

__email__ = "michael.hellmig@helmholtz-berlin.de"
__status__ = "Beta"


TOOLS = {"FLANGE": 0, "CRYOTONG": 1, "ESRF": 2, "PLATE": 3}


class SOLEILCatsMaint(HardwareObject):
    __TYPE__ = "CATS"

    NO_OF_LIDS = 3

    default_tool = 1
    default_soak_lid = 2

    """
    Actual implementation of the CATS Sample Changer, MAINTENANCE COMMANDS ONLY
    BESSY BL14.1 installation with 3 lids
    """

    def __init__(self, *args, **kwargs):
        logging.info("CatsMaint: __init__")
        super().__init__(*args, **kwargs)

    def init(self):
        logging.info("CatsMaint: init")

        tool = self.get_property("tool")
        if tool in TOOLS:
            self.tool = TOOLS[tool]
        else:
            self.tool = self.default_tool

        soaklid = self.get_property("soak_lid")
        if soaklid is not None:
            self.soaklid = soaklid
        else:
            self.soaklid = self.default_soak_lid

        self.running_safe = False

        self._chnPathRunning = self.get_channel_object("_chnPathRunning")
        self._chnPathRunning.connect_signal("update", self._update_running_state)
        self._chnPowered = self.get_channel_object("_chnPowered")
        self._chnPowered.connect_signal("update", self._update_powered_state)
        self._chnMessage = self.get_channel_object("_chnMessage")
        self._chnMessage.connect_signal("update", self._update_message)
        self._chnLN2Regulation = self.get_channel_object("_chnLN2RegulationDewar1")
        self._chnLN2Regulation.connect_signal("update", self._update_regulation_state)

        for command_name in (
            "_cmdResetError",
            "_cmdBack",
            "_cmdSafe",
            "_cmdHome",
            "_cmdDry",
            "_cmdDrySoak",
            "_cmdSoak",
            "_cmdClearMemory",
            "_cmdAckSampleMemory",
            "_cmdOpenTool",
            "_cmdToolCal",
            "_cmdPowerOn",
            "_cmdPowerOff",
            "_cmdOpenLid1",
            "_cmdCloseLid1",
            "_cmdOpenLid2",
            "_cmdCloseLid2",
            "_cmdOpenLid3",
            "_cmdCloseLid3",
            "_cmdRegulOn",
        ):
            setattr(self, command_name, self.get_command_object(command_name))

        for lid_index in range(self.NO_OF_LIDS):
            channel_name = "_chnLid%dState" % (lid_index + 1)
            setattr(self, channel_name, self.get_channel_object(channel_name))
            if getattr(self, channel_name) is not None:
                getattr(self, channel_name).connect_signal(
                    "update", getattr(self, "_updateLid%dState" % (lid_index + 1))
                )

    ################################################################################

    def back_traj(self):
        """
        Moves a sample from the gripper back into the dewar to its logged position.
        """
        return self._execute_task(False, self._do_back)

    def safe_traj(self):
        """
        Safely Moves the robot arm and the gripper to the home position
        """
        return self._execute_task(False, self._do_safe)

    # MS 2014-11-18
    def home_traj(self):
        """
        Moves the robot arm to the home position
        """
        return self._execute_task(False, self._do_home)

    def dry_traj(self):
        """
        Drying the gripper
        """
        return self._execute_task(False, self._do_dry)

    def dry_soak_traj(self):
        """
        Dry and Soak the gripper
        """
        return self._execute_task(False, self._do_dry_soak)

    def soak_traj(self):
        """
        Soaking the gripper
        """
        return self._execute_task(False, self._do_soak)

    def integrated_tool_cal(self):
        return self._execute_task(False, self._do_integrated_tool_cal)

    def clear_memory(self):
        """
        Clears the memory
        """
        return self._execute_task(False, self._do_clear_memory)

    def ack_sample_memory(self):
        """
        Acknowledge incoherence between memorized and actual sample status -- e.g. if robot executed put trajectory but no sample was mounted on the gonio -- either because of empty position or problem with gripper.
        """
        return self._execute_task(False, self._do_ack_sample_memory)

    def opentool(self):
        """
        Drying the gripper
        """
        return self._execute_task(False, self._do_opentool)

    def toolcal_traj(self):
        """
        Soaking the gripper
        """
        return self._execute_task(False, self._do_tool_cal)

    ###

    def missingSample(self):
        self._do_ack_sample_memory()
        self._do_clear_memory()
        self._do_resetError()

    def _do_abort(self):
        """
        Launch the "abort" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdAbort()

    def _do_resetError(self):
        """
        Launch the "reset" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdResetError()

    def _do_back(self):
        """
        Launch the "back" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = self.tool
        self._execute_server_task(self._cmdBack, argin)

    def _do_safe(self):
        """
        Launch the "safe" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = self.tool
        self._execute_server_task(self._cmdSafe, argin)

    # MS 2014-11-18
    def _do_home(self):
        """
        Launch the "home" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = self.tool
        self._execute_server_task(self._cmdHome, argin)

    def _do_dry(self):
        """
        Launch the "dry" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = (str(self.tool),)
        self._execute_server_task(self._cmdDry, argin)

    def _do_dry_gripper(self):
        return self._do_dry_soak()

    def _do_dry_soak(self):
        """
        Launch the "dry_soak" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = map(str, (self.tool, self.soaklid))
        self._execute_server_task(self._cmdDrySoak, argin)

    def _do_soak(self):
        """
        Launch the "soak" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = map(str, (self.tool, self.soaklid))
        self._execute_server_task(self._cmdSoak, argin)

    def _do_integrated_tool_cal(self):
        argin = self.tool

        # poweron
        # safe
        # dry
        # tremp_toolcal
        # soak

        self.running_safe = True
        self._update_message("RUNNING Safe process. Phase is: going to Safe")
        self.log.info("Executing safe")
        self._execute_server_task(self._cmdSafe, argin, waitstart=True)
        self.log.info("Executing safe done")

        self._update_message("RUNNING Safe process. Phase is: Drying")
        self.log.info("Executing dry")
        self._execute_server_task(self._cmdDry, argin, waitstart=True)
        self.log.info("Executing dry done")

        self._update_message("RUNNING Safe process. Phase is: SoakToolCal")
        logging.getLogger("Executing SoakToolCal")
        self._execute_server_task(self._cmdSoakToolCal, argin, waitstart=True)
        logging.getLogger("Executing SoakToolCal DONE")

        self._update_message("RUNNING Safe process. Phase is: ResetError")
        self._cmdResetError()

        logging.getLogger("Executing SoakToolCal")
        self._update_message(
            "RUNNING Safe process. Phase is: Soak (will take 45 seconds)"
        )
        argin = map(str, (self.tool, self.soaklid))
        self._execute_server_task(self._cmdSoak, argin, waitstart=True)
        logging.getLogger("Executing SoakToolCal done")
        self._update_message("Safe process finished")
        self.running_safe = False

    def _do_clear_memory(self):
        """
        Execute "clear_memory" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        # argin = 1
        self._execute_server_task(self._cmdClearMemory)

    def _do_ack_sample_memory(self):
        """
        Execute "clear_memory" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        # argin = 1
        self._execute_server_task(self._cmdAckSampleMemory)

    def _do_opentool(self):
        """
        Open tool via the CATS Tango DS

        :returns: None
        :rtype: None
        """
        # argin = 1
        self._execute_server_task(self._cmdOpenTool)  # , argin)

    def _do_tool_cal(self):
        """
        Launch the "toolcal" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        # argin = 1
        self._execute_server_task(self._cmdToolCal)

    ###

    def _do_power_state(self, state=False):
        """
        Switch on CATS power if >state< == True, power off otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._cmdPowerOn()
        else:
            self._cmdPowerOff()

    def _do_enable_regulation(self):
        """
        Switch on CATS regulation

        :returns: None
        :rtype: None
        """
        self._cmdRegulOn()

    def _do_lid1_state(self, state=True):
        """
        Opens lid 1 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._execute_server_task(self._cmdOpenLid1)
        else:
            self._execute_server_task(self._cmdCloseLid1)

    def _do_lid2_state(self, state=True):
        """
        Opens lid 2 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._execute_server_task(self._cmdOpenLid2)
        else:
            self._execute_server_task(self._cmdCloseLid2)

    def _do_lid3_state(self, state=True):
        """
        Opens lid 3 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._execute_server_task(self._cmdOpenLid3)
        else:
            self._execute_server_task(self._cmdCloseLid3)

    #########################          PROTECTED          #########################

    def _execute_task(self, wait, method, *args):
        ret = self._run(method, wait=False, *args)
        if wait:
            return ret.get()
        else:
            return ret

    @task
    def _run(self, method, *args):
        exception = None
        ret = None
        try:
            ret = method(*args)
        except Exception as ex:
            exception = ex
        if exception is not None:
            raise exception
        return ret

    #########################           PRIVATE           #########################

    def _update_running_state(self, value):
        logging.info("CatsMaint _update_running_state %s" % value)
        self.emit("runningStateChanged", (value,))

    def _update_powered_state(self, value):
        logging.info("CatsMaint _update_powered_state %s" % value)
        self.emit("powerStateChanged", (value,))

    def _update_message(self, value):
        logging.info("CatsMaint _update_message %s" % value)
        if value.strip() == "" and self.running_safe:
            return

        if "incoherent" in value.lower():
            value = (
                '%s\nThe sample is not present on the gonio although robot thinks it should be.\nThis can happen in three cases:\n1. there was no sample in the specified position in the puck,\n2. the robot could not get it (rare)\n3. the gonio can not detect sample which is present (very rare).\nIf the sample is really not present on the gonio please click "abort" button\n and then "Missing sample" button below to be able to continue.'
                % value
            )
        if "trfgtd" in value.lower():
            value = (
                "%s\nTransfer permission was not granted by the gonio.\n1. Please Abort the trajectory\n2. set gonio to Transfer phase from the pull down menu on the right\n3. Start the load/unload trajectory again."
                % value
            )
        if "dback" in value.lower():
            value = (
                "%s\nThe detector is too close (less then 180 mm from the sample)\nPlease move it to at least 181 mm."
                % value
            )
        if "remote mode requested" in value.lower():
            value = (
                "%s\nRemote operation not enabled.\nPlease turn the robot key to the remote position.\nThe key is located next to the experiment hutch door."
                % value
            )
        self.emit("messageChanged", (value,))

    def _update_regulation_state(self, value):
        logging.info("CatsMaint _update_regulation_state %s" % value)
        self.emit("regulationStateChanged", (value,))

    def _update_lid1_state(self, value):
        logging.info("CatsMaint _update_lid1_state %s" % value)
        self.emit("lid1StateChanged", (value,))

    def _update_lid2_state(self, value):
        logging.info("CatsMaint _update_lid2_state %s" % value)
        self.emit("lid2StateChanged", (value,))

    def _update_lid3_state(self, value):
        logging.info("CatsMaint _update_lid3_state %s" % value)
        self.emit("lid3StateChanged", (value,))

    def _update_operation_mode(self, value):
        self._scIsCharging = not value

    def _execute_server_task(self, method, argin=None, waitstart=False):
        if argin is not None:
            task_id = method(argin)
        else:
            task_id = method()

        print("CatsMaint._execute_server_task", task_id)
        ret = None
        # introduced wait because it takes some time before the attribute PathRunning is set
        # after launching a transfer
        # after setting refresh in the Tango DS to 0.1 s a wait of 1s is enough

        if waitstart:
            timeout = 10.0
            t0 = time.time()
            while str(self._chnPathRunning.get_value()).lower() != "true":
                if time.time() - t0 > timeout:
                    self.log.info("Could not detect the start of the task. Continuing")
                    break
                gevent.sleep(0.1)

        self.log.info(
            "server task started. path running %s" % self._chnPathRunning.get_value()
        )
        while str(self._chnPathRunning.get_value()).lower() == "true":
            gevent.sleep(0.1)
        ret = True
        return ret



# encoding: utf-8
#
# This file is part of MXCuBE.
#
# MXCuBE is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# MXCuBE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MXCuBE.  If not, see <https://www.gnu.org/licenses/>.
""" """

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import logging

from mxcubecore.HardwareObjects.abstract.AbstractXrayCentring import (
    AbstractXrayCentring,
)
from mxcubecore.model import queue_model_objects

__copyright__ = """ Copyright © 2016 - 2022 by MXCuBE Collaboration """
__license__ = "LGPLv3+"
__author__ = "rhfogh"
__date__ = "25/03/2022"
__category__ = "General"


class XrayCentringMockup(AbstractXrayCentring):
    def execute(self):
        """Dummy execution. Adds new entries to queue, to serve as example."""

        queue_manager = self._queue_entry.get_queue_controller()

        logging.getLogger("HWR").debug("Execute mock Xray centring ")

        # Add Dummy task 1
        dummy = queue_model_objects.DelayTask(delay=1)
        dummy.set_enabled(True)
        dummy.set_name("CentringDelay1")
        self._add_to_queue(self._data_collection_group, dummy)
        dummy_entry = queue_manager.get_entry_with_model(dummy)
        dummy_entry.in_queue = True

        # Add Dummy task 2
        dummy = queue_model_objects.DelayTask(delay=2)
        dummy.set_enabled(True)
        dummy.set_name("CentringDelay2")
        self._add_to_queue(self._data_collection_group, dummy)
        dummy_entry = queue_manager.get_entry_with_model(dummy)
        dummy_entry.in_queue = True

        # Execute
        data_collection_entry = queue_manager.get_entry_with_model(
            self._data_collection_group
        )
        queue_manager.execute_entry(data_collection_entry)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""AbstractNState class - interface for N state devices.
Defines BaseValueEnum, initialise_values and value_to_enum methods.
Implements validate_value, set/update limits.
"""

import abc
from ast import literal_eval
from enum import (
    Enum,
    unique,
)

from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@unique
class BaseValueEnum(Enum):
    """Defines only the compulsory unknown."""

    UNKNOWN = "UNKNOWN"


class AbstractNState(AbstractActuator):
    """Abstract base class for N state objects."""

    __metaclass__ = abc.ABCMeta
    VALUES = BaseValueEnum

    def init(self):
        """Initialise the predefined values"""
        super().init()
        self.initialise_values()
        # default_value should be an Enum, if defined
        if self.default_value is not None and not isinstance(self.default_value, Enum):
            self.default_value = self.value_to_enum(self.default_value)
            self.update_value(self.default_value)

    def validate_value(self, value):
        """Check if the value is one of the predefined values.
        Args:
            value(Enum): value to check
        Returns:
            (bool): True if within the values.
        """
        return value in self.VALUES

    def set_limits(self, limits):
        """Set the low and high limits.
        Args:
            limits (tuple): two element (low limit, high limit) tuple.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def update_limits(self, limits=None):
        """Check if the limits have changed.
        Args:
            limits(tuple): two elements (low limit, high limit) tuple.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def initialise_values(self):
        """Initialise the ValueEnum with the values from the config."""
        values = self.get_property("values", [])
        if isinstance(values, str):
            values = literal_eval(values)
        values_dict = dict(**{item.name: item.value for item in self.VALUES})
        values_dict.update(values)
        self.VALUES = Enum("ValueEnum", values_dict)

    def value_to_enum(self, value, idx=0):
        """Transform a value to Enum
        Args:
           value(str, int, float, tuple): value
           idx(int): Index in the value, if tuple
        Returns:
            (Enum): Enum member, corresponding to the value or UNKNOWN.
        """
        for enum_var in self.VALUES:
            if value == enum_var.value:
                return enum_var
            if isinstance(enum_var.value, tuple) and value == enum_var.value[idx]:
                return enum_var

        return self.VALUES.UNKNOWN

    def re_emit_values(self):
        """Update values for all internal attributes"""
        self.update_value(self.get_value())

        # NB DO NOT 'FIX', this is deliberate.
        # One would normally call super(AbstractNState ...), however we
        # want to call re_emit_values of HardwareObject to avoid the limit
        # handling implemented in AbstractActuator
        super(AbstractActuator, self).re_emit_values()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Generic ESRF Beam Definer.

Example xml configuration:

.. code-block:: xml

 <object class="ESRF.ESRFBeamDefiner"
   <username>Beam Definer</username>
   <beam_config>
      <name>4x4 um</name>
      <beam_size>0.004, 0.004</beam_size>
   </beam_config>
   <beam_config>
      <name>4x8 um</name>
      <beam_size>0.004, 0.008</beam_size>
   </beam_config>
   <default_size_name>4x4 um</default_size_name>
 </object>
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"

from ast import literal_eval
from enum import Enum

from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState


class ESRFBeamDefiner(AbstractNState):
    """Generic ESRF beam definer implementation"""

    def __init__(self, *args):
        super().__init__(*args)
        self.beam_config = {}
        self.bd_config = []

    def init(self):
        super().init()
        self._default_name = self.get_property("default_size_name")

        # keep the config is needed by the inheriring classes
        self.bd_config = self.init_config()

        # check if we have values other that UNKNOWN
        if len(self.VALUES) == 1:
            self._initialise_values()

    def init_config(self):
        """Get the configutarion from the file"""

        cfg = self.get_property("beam_config")
        if not isinstance(cfg, list):
            cfg = [cfg]

        for beam_cfg in cfg:
            name = beam_cfg.get_property("name")
            beam_size = beam_cfg.get_property("beam_size", (0.015, 0.015))
            if isinstance(beam_size, str):
                beam_size = literal_eval(beam_size)
            self.beam_config.update({name: beam_size})
        return cfg

    def get_limits(self):
        return (1, len(self.beam_config))

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): The current position Enum.
        """
        try:
            return self.VALUES[self.get_current_position_name()]
        except (ValueError, KeyError):
            return self.VALUES.UNKNOWN

    def get_size(self):
        """Get the current beam size (horizontal and vertical).
        Returns:
            (tuple): Current beam size (horizontal, vertical) [mm].
        """
        return self.get_value().value

    def get_current_position_name(self):
        """Get the current beam size name.
        Returns:
            (str): Current beam size name.
        """
        raise NotImplementedError

    def get_predefined_positions_list(self):
        """Get the list of all the beam size names.
        Returns:
            (list): List of strings with the beam size names.
        """
        return list(self.beam_config.keys())

    def _initialise_values(self):
        """Initialise the ValueEnum from the configuration.
        Raises:
            RuntimeError: No values defined.
        """
        self.VALUES = Enum(
            "ValueEnum",
            dict(self.beam_config, **{item.name: item.value for item in self.VALUES}),
        )



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import time

import gevent
from XSDataAutoprocv1_0 import XSDataAutoprocInput
from XSDataCommon import (
    XSDataDouble,
    XSDataFile,
    XSDataInteger,
    XSDataString,
)

from mxcubecore.BaseHardwareObjects import HardwareObject


class MAXIVAutoProcessing(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)
        self.result = None
        self.autoproc_programs = None
        self.current_autoproc_procedure = None

    def init(self):
        self.autoproc_programs = self["programs"]

    def execute_autoprocessing(self, process_event, params_dict, frame_number):
        if self.autoproc_programs is not None:
            self.current_autoproc_procedure = gevent.spawn(
                self.autoproc_procedure, process_event, params_dict, frame_number
            )
            self.current_autoproc_procedure.link(self.autoproc_done)

    def autoproc_procedure(self, process_event, params_dict, frame_number):
        """Main autoprocessing procedure.

        At the beginning correct event (defined in xml) is found. If the event
        is executable then accordingly to the event type (image, after) then
        the sequence is executed:

        Implemented tasks:
           - after : Main autoprocessing procedure
                     1. Input file is generated with create_autoproc_input
                        Input file has name "edna-autoproc-input-%Y%m%d_%H%M%S.xml".
                     2. Then it waits for XDS.INP directory and if it exists then
                        creates input file
                     3. edna_autoprocessing.sh script is executed with parameters:
                        - arg1 : generated xml file
                        - arg2 : process dir
                     4. script executes EDNA EDPluginControlAutoprocv1_0
           - image : Thumbnail generation for first and last image
                     1. No input file is generated
                     2. edna_thumbnails.sh script is executed with parameters:
                        - arg1 : image base dir (place where thumb will be generated)
                        - arg2 : file name
        """
        for program in self.autoproc_programs:
            if process_event == program.get_property("event"):
                module = program.get_property("module").lower()
                print(2 * "###########")
                if process_event == "after":
                    input_filename, will_execute = self.create_autoproc_input(
                        process_event, params_dict
                    )
                    path = params_dict["auto_dir"]
                    mode = "after"
                    dataCollectionId = str(params_dict["collection_id"])
                    residues = 200
                    anomalous = False
                    cell = "0,0,0,0,0,0"
                    print("Module: ", module, ">> Will execute: ", will_execute)
                    print("input_filename   ", input_filename)
                    try:
                        if module == "ednaproc" and will_execute:
                            from ednaProcLauncher import EdnaProcLauncher

                            mod = EdnaProcLauncher(
                                path,
                                mode,
                                dataCollectionId,
                                residues,
                                anomalous,
                                cell,
                                None,
                            )

                        elif module == "autoproc" and will_execute:
                            from autoProcLauncher import AutoProcLauncher

                            mod = AutoProcLauncher(
                                path,
                                mode,
                                dataCollectionId,
                                residues,
                                anomalous,
                                cell,
                                None,
                            )
                    except Exception as ex:
                        print(ex)
                if process_event == "image":
                    if (
                        frame_number == 1
                        or frame_number
                        == params_dict["oscillation_sequence"][0]["number_of_images"]
                    ):
                        endOfLineToExecute = " %s %s/%s_%d_%05d.cbf" % (
                            params_dict["fileinfo"]["directory"],
                            params_dict["fileinfo"]["directory"],
                            params_dict["fileinfo"]["prefix"],
                            params_dict["fileinfo"]["run_number"],
                            frame_number,
                        )
                        will_execute = True
                if will_execute:
                    self.log.info("[MAXIVAutoprocessing] Executing module: %s" % module)
                    try:
                        mod.parse_and_execute()
                    except Exception as ex:
                        self.log.error(
                            "[MAXIVAutoprocessing] Module %s  execution error." % module
                        )
                        print(module, ex)

    def autoproc_done(self, current_autoproc):
        self.current_autoproc_procedure = None
        self.log.info("Autoprocessing executed.")

    def create_autoproc_input(self, event, params):
        WAIT_XDS_TIMEOUT = 20
        WAIT_XDS_RESOLUTION = 1

        file_name_timestamp = time.strftime("%Y%m%d_%H%M%S")

        autoproc_path = params.get("xds_dir")
        autoproc_xds_filename = os.path.join(autoproc_path, "XDS.INP")
        autoproc_input_filename = os.path.join(
            autoproc_path, "edna-autoproc-input-%s" % file_name_timestamp
        )
        autoproc_output_file_name = os.path.join(
            autoproc_path, "edna-autoproc-results-%s" % file_name_timestamp
        )

        autoproc_input = XSDataAutoprocInput()
        autoproc_xds_file = XSDataFile()
        autoproc_xds_file.setPath(XSDataString(autoproc_xds_filename))
        autoproc_input.setInput_file(autoproc_xds_file)

        autoproc_output_file = XSDataFile()
        autoproc_output_file.setPath(XSDataString(autoproc_output_file_name))
        autoproc_input.setOutput_file(autoproc_output_file)

        autoproc_input.setData_collection_id(XSDataInteger(params.get("collection_id")))
        residues_num = float(params.get("residues", 0))
        if residues_num != 0:
            autoproc_input.setNres(XSDataDouble(residues_num))
        space_group = params.get("sample_reference").get("spacegroup", "")
        if not isinstance(space_group, int) and len(space_group) > 0:
            autoproc_input.setSpacegroup(XSDataString(space_group))
        unit_cell = params.get("sample_reference").get("cell", "")
        if len(unit_cell) > 0:
            autoproc_input.setUnit_cell(XSDataString(unit_cell))

        autoproc_input.setCc_half_cutoff(XSDataDouble(18.0))

        # Maybe we have to check if directory is there. Maybe create dir with mxcube
        xds_appeared = False
        wait_xds_start = time.time()
        logging.info(
            "MAXIVAutoprocessing: Waiting for XDS.INP file: %s" % autoproc_xds_filename
        )
        while not xds_appeared and time.time() - wait_xds_start < WAIT_XDS_TIMEOUT:
            if (
                os.path.exists(autoproc_xds_filename)
                and os.stat(autoproc_xds_filename).st_size > 0
            ):
                xds_appeared = True
                logging.debug(
                    "MAXIVAutoprocessing: XDS.INP file is there, size={0}".format(
                        os.stat(autoproc_xds_filename).st_size
                    )
                )
            else:
                os.system("ls %s> /dev/null" % (os.path.dirname(autoproc_path)))
                gevent.sleep(WAIT_XDS_RESOLUTION)
        if not xds_appeared:
            logging.error(
                "MAXIVAutoprocessing: XDS.INP file ({0}) failed to appear after {1} seconds".format(
                    autoproc_xds_filename, WAIT_XDS_TIMEOUT
                )
            )
            return None, False

        autoproc_input.exportToFile(autoproc_input_filename)

        return autoproc_input_filename, True



import time

import gevent
import numpy as np

from mxcubecore.HardwareObjects.abstract.AbstractMotor import MotorStates
from mxcubecore.HardwareObjects.ExporterMotor import ExporterMotor


class MicrodiffKappaMotor(ExporterMotor):
    lock = gevent.lock.Semaphore()
    motors = {}
    conf = {}

    def __init__(self, name):
        ExporterMotor.__init__(self, name)

    def init(self):
        ExporterMotor.init(self)
        if not self.actuator_name in ("Kappa", "Phi"):
            raise RuntimeError("MicrodiffKappaMotor class is only for kappa motors")
        MicrodiffKappaMotor.motors[self.actuator_name] = self
        if self.actuator_name == "Kappa":
            MicrodiffKappaMotor.conf["KappaTrans"] = self.stringToList(
                self.get_property("kappaTrans")
            )
            MicrodiffKappaMotor.conf["KappaTransD"] = self.stringToList(
                self.get_property("kappaTransD")
            )
        elif self.actuator_name == "Phi":
            MicrodiffKappaMotor.conf["PhiTrans"] = self.stringToList(
                self.get_property("phiTrans")
            )
            MicrodiffKappaMotor.conf["PhiTransD"] = self.stringToList(
                self.get_property("phiTransD")
            )
        self.sampx = self.get_object_by_role("sampx")
        self.sampy = self.get_object_by_role("sampy")
        self.phiy = self.get_object_by_role("phiy")

    def stringToList(self, commaSeparatedString):
        return [float(x) for x in commaSeparatedString.split(",")]

    def move(self, absolutePosition):
        kappa_start_pos = MicrodiffKappaMotor.motors["Kappa"].get_value()
        kappa_phi_start_pos = MicrodiffKappaMotor.motors["Phi"].get_value()
        if self.actuator_name == "Kappa":
            kappa_end_pos = absolutePosition
            kappa_phi_end_pos = kappa_phi_start_pos
        else:
            kappa_end_pos = kappa_start_pos
            kappa_phi_end_pos = absolutePosition
        sampx_start_pos = self.sampx.get_value()
        sampy_start_pos = self.sampy.get_value()
        phiy_start_pos = self.phiy.get_value()
        """
        with MicrodiffKappaMotor.lock:
            if self.get_state() != MotorStates.UNKNOWN:
                self.position_attr.set_value(
                    absolutePosition
                )  # absolutePosition-self.offset)
                self.update_state(MotorStates.MOVING)

            # calculations
            newSamplePositions = self.getNewSamplePosition(
                kappa_start_pos,
                kappa_phi_start_pos,
                sampx_start_pos,
                sampy_start_pos,
                phiy_start_pos,
                kappa_end_pos,
                kappa_phi_end_pos,
            )
            self.sampx.set_value(newSamplePositions["sampx"])
            self.sampy.set_value(newSamplePositions["sampy"])
            self.phiy.set_value(newSamplePositions["phiy"])
        """

    def waitEndOfMove(self, timeout=None):
        with gevent.Timeout(timeout):
            time.sleep(0.1)
            while self.motorState == MotorStates.MOVING:
                time.sleep(0.1)
            self.sampx.waitEndOfMove()
            self.sampy.waitEndOfMove()
            self.phiy.waitEndOfMove()

    def stop(self):
        if self.get_state() != MotorStates.NOTINITIALIZED:
            self._motor_abort()
        for m in (self.sampx, self.sampy, self.phiy):
            m.stop()

    def getNewSamplePosition(
        self, kappaAngle1, phiAngle1, sampx, sampy, phiy, kappaAngle2, phiAngle2
    ):
        """
        This method calculates the translation correction for inversed kappa goniostats.
        For more info see Acta Cryst.(2011). A67, 219-228, Sandor Brockhauser et al., formula (3).
        See also MXSUP-1823.
        """
        self.log.info("In MicrodiffKappaMotor.getNewSamplePosition")
        self.log.info(
            "Input arguments: Kappa %.2f Phi %.2f sampx %.3f sampy %.3f phiy %.3f Kappa2 %.2f Phi2 %.2f"
            % (kappaAngle1, phiAngle1, sampx, sampy, phiy, kappaAngle2, phiAngle2)
        )
        kappaRot = np.array(MicrodiffKappaMotor.conf["KappaTransD"])
        phiRot = np.array(MicrodiffKappaMotor.conf["PhiTransD"])
        t_kappa_zero = np.array(MicrodiffKappaMotor.conf["KappaTrans"])
        t_phi_zero = np.array(MicrodiffKappaMotor.conf["PhiTrans"])
        t_start = np.array([-sampx, -sampy, -phiy])
        # if beamline in ["id29", "id30b"]:
        #    t_start = np.array([-sampx, -sampy, -phiy])
        # else:
        #    t_start = np.array([sampx, sampy, -phiy])
        kappaRotMat1 = self.rotation_matrix(kappaRot, -kappaAngle1 * np.pi / 180.0)
        kappaRotMat2 = self.rotation_matrix(kappaRot, kappaAngle2 * np.pi / 180.0)
        phiRotMat = self.rotation_matrix(
            phiRot, (phiAngle2 - phiAngle1) * np.pi / 180.0
        )
        t_step1 = t_kappa_zero - t_start
        t_step2 = t_kappa_zero - np.dot(kappaRotMat1, t_step1)
        t_step3 = t_phi_zero - t_step2
        t_step4 = t_phi_zero - np.dot(phiRotMat, t_step3)
        t_step5 = t_kappa_zero - t_step4
        t_end = t_kappa_zero - np.dot(kappaRotMat2, t_step5)
        new_motor_pos = {}
        new_motor_pos["sampx"] = float(-t_end[0])
        new_motor_pos["sampy"] = float(-t_end[1])
        new_motor_pos["phiy"] = float(-t_end[2])
        self.log.info("New motor positions: %r" % new_motor_pos)
        return new_motor_pos

    def rotation_invariant(self, v):
        return np.outer(v, v)

    def skew_symmetric(self, v):
        l, m, n = v
        return np.array([[0, -n, m], [n, 0, -l], [-m, l, 0]])

    def inverse_skew_symmetric(self, v):
        l, m, n = v
        return np.array([[0, n, -m], [-n, 0, l], [m, -l, 0]])

    def rotation_symmetric(self, v):
        return np.identity(3) - np.outer(v, v)

    def rotation_matrix(self, axis, theta):
        return (
            self.rotation_invariant(axis)
            + self.skew_symmetric(axis) * np.sin(theta)
            + self.rotation_symmetric(axis) * np.cos(theta)
        )



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract Diffractometer class.
Initialises the username property and all the motors and nstate (discrete
positions) equipment, which are part of the diffractometer.
Certain number of roles are fixed and define a corresponding motor or nstate
actuator object. This allows to use them in a standard way by the MXCuBE
application (web or Qt).
There is also a convention of the direction of the motors:
  - x axis is parallel to the beam. Positive direction is from right to left.
  - y axis is perpendicular to the beam, parallel to the ground. Positive
      direction is from left to right when facing the beam.
  - z axis is perpendicular to the floor. Positive direction is top down.

Here follows the list of the fixed roles and the description of the
corresponding objects, accessible via beamline.diffractometer hardware object.

1. Motor objects (roles)  and their functionality:
  omega - the rotation axis, independent of the orientation (up, down or side).
          Pisitive direction is clockwise.
  sampx - centring table x axis
  sampy - centring table y axis
  focus - alignment table x axis
  phiy - alignment table y axis
  phiz - alignment table z axis
  sample_horizontal - x axis, combination of sampx and sampy. Equivalent to
                      sampx at omega=0 and sampy for omega=90.
  samle_vertical - y axis, combination of sampx and sampy. Equivalent to
                   sampy at omega=0 and sampx for omega=90.
  backlight - adjust the intensity of the back light
  frontlight - adjust the intensity of the front light
  kappa - the minikappa kappa axis
  kappa_phi - the minikappa phi axis

2. Discrete (N) state equipment and its functionality:
zoom - zoom levels
fshutter - fast (ms) shutter - allow beam on the sample
beamstop - put in front of a detector to avoid the direct beam.
capillary - if present, a tube to reduce the scattering background.
backlightswitch - move the backlight on the level of the on-axis viewer
frontlightswitch - switch on/off the front light
fluo_detector - if present, actuator to move a fluorescence detector close to the sample
"""

import abc
import json
from enum import Enum, unique
from typing import Dict, List, Optional, Tuple, Union
from pathlib import Path
from pydantic.v1 import BaseModel, Field, ValidationError

from mxcubecore.BaseHardwareObjects import HardwareObject, HardwareObjectState
from mxcubecore import HardwareRepository as HWR


__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@unique
class DiffractometerHead(Enum):
    """Enumeration diffractometer head types."""

    UNKNOWN = "Unknown"
    MINI_KAPPA = "MiniKappa"
    SMART_MAGNET = "SmartMagnet"
    PLATE = "Plate"
    SSX = "SSX"
    HARVESTER = "Harvester"


@unique
class DiffractometerPhase(Enum):
    """Enumeration diffractometer phases."""

    UNKNOWN = "Unknown"
    CENTRE = "Centring"
    COLLECT = "DataCollection"
    SEE_BEAM = "BeamLocation"
    TRANSFER = "Transfer"


@unique
class DiffractometerConstraint(Enum):
    """Enumeration diffractometer constraint types."""

    UNKNOWN = "Unknown"
    RELEASE = "Normal"
    INJECTOR = "Injector"
    STILL = "LockRotation"


class HolderTypeEnum(Enum):
    """Enumeration of chip holder geometry."""

    KNOWN_GEOMETRY = "known_geometry"
    FREE_GEOMETRY = "free_geometry"


class ChipShapeEnum(Enum):
    """Enumeration of chip holder shape."""

    RECTANGULAR = "RECTANGULAR"
    ELLIPTICAL = "ELLIPTICAL"


# Diffractometer pydantic models


class CalibrationData(BaseModel):
    """Chip calibration model."""

    top_left: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Top left corner motor position"
    )
    top_right: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Top right corner motor position"
    )
    bottom_left: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Bottom left corner motor position"
    )


class SampleHolderSectionModel(BaseModel):
    """Generic sample holder."""

    calibration_data: Optional[CalibrationData]
    section_offset: Tuple[int, int] = Field(
        [0, 0], description="Block offset in grid layout system coordinates x, y"
    )
    block_size: Tuple[float, float] = Field(
        [15, 15], description="Block size horizontal, vertical in mm"
    )
    block_spacing: Tuple[float, float] = Field(
        [15, 15], description="Spacing between blocks horizontal, vertical in mm"
    )
    block_shape: ChipShapeEnum = ChipShapeEnum.RECTANGULAR
    number_of_rows: int = Field(6, description="Numer of rows")
    number_of_collumns: int = Field(6, description="Numer of collumns")
    row_labels: List[str] = Field([], description="Row lables")
    column_lables: List[str] = Field([], description="Collumn lables")
    targets_per_block: Tuple[int, int] = Field(
        [20, 20], description="Targets per block dim1 and dim2"
    )


class ChipLayout(BaseModel):
    """Chip layout model."""

    head_type: DiffractometerHead = DiffractometerHead.SSX
    holder_type: HolderTypeEnum = HolderTypeEnum.KNOWN_GEOMETRY
    holder_brand: str = Field("", description="Brand/make of sample holder")
    holder_size: Tuple[float, float] = Field(
        [0, 0], description="Size of sample holder in mm horizontal and vertical"
    )
    sections: List[SampleHolderSectionModel]
    calibration_data: CalibrationData


class GonioHeadConfiguration(BaseModel):
    current: str = Field("", description="Selected chip layout")
    available: Dict[str, ChipLayout]


class AbstractDiffractometer(HardwareObject):
    """Abstract Diffractometer.

    Attributes:
        motors_hwobj_dict (dict): Motor hardware objects.
        nstate_equipment_hwobj_dict (dict): N state hardware objects.
        username (str): User name
        head_type (DiffractometerHead Enum): Current head type
        current_phase (DiffractometerPhase Enum): Current phase
        current_constraint (DiffractometerConstraint Enum): Current constraint.
        timeout (float): Default action timeout [s]

    Emits:
        valueChanged: ("valueChanged", (value,))
        phaseChanged: ("phaseChanged", (state))

    States:
        HardwareObjectStates: READY, BUSY, FAULT
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self.motors_hwobj_dict = {}
        self.nstate_equipment_hwobj_dict = {}
        self.username = name
        self.current_phase = None
        self.head_type = None
        self.current_constraint = None
        self.timeout = 3  # default timeout 3 s
        self.chip_definition_file = ""

    def init(self):
        """Initialise username property.
        Initialise the equipment, defined in the configuration file
        """
        self.username = self.get_property("username") or self.username

        # motors
        for role in self.config.motors:
            try:
                _hobj = self.get_object_by_role(role)
                self.motors_hwobj_dict[role] = _hobj
                setattr(self, role, _hobj)
                self.connect(_hobj, "valueChanged", _hobj.update_value)
            except KeyError:
                self.log.warning("Diffractometer: No motors configured")

        # nstate (discrete positions) equipment
        for role in self.config.nstate_equipment:
            try:
                _hobj = self.get_object_by_role(role)
                self.nstate_equipment_hwobj_dict[role] = _hobj
                setattr(self, role, _hobj)
                self.connect(_hobj, "valueChanged", _hobj.update_value)
            except KeyError:
                self.log.warning("No nstate (discrete positions) equipment configured")

        # chip definition
        _fp = self.get_property("chip_definition_file", "")
        self.chip_definition_file = HWR.get_hardware_repository().find_in_repository(
            _fp
        )

    def get_motors(self) -> dict:
        """Get the dictionary of all configured motors or the ones to use.

        Returns:
            Dictionary {role: hardware_object}
        """
        return self.motors_hwobj_dict.copy()

    def get_nstate_equipment(self) -> dict:
        """Get the dictionary of all the nstate (discrete positions) equipment.

        Returns:
            Dictionary {role: hardware_object}
        """
        return self.nstate_equipment_hwobj_dict.copy()

    # -------- Motor Groups --------

    def set_value_motors(
        self,
        motors_positions_dict: dict,
        simultaneous: bool = True,
        timeout: float | None = None,
    ):
        """Move specified motors to the requested positions

        Args:
            motors_positions_dict: Dictionary {motor_role: target_value}.
            simultaneous: Move the motors simultaneousl (True - default) or not.
            timeout: timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            TimeoutError: Timeout
            KeyError: The name does not correspond to an existing motor
        """

        # use only the available motors
        mot_hwobj_dict = self.get_motors()

        tout = timeout
        if simultaneous:
            tout = 0

        self.update_state(HardwareObjectState.BUSY)
        for key, val in motors_positions_dict.items():
            try:
                mot_hwobj_dict[key].set_value(val, timeout=tout)
            except KeyError as err:
                msg = f"Invalid motor name {key}"
                raise RuntimeError(msg) from err

        # wait for the end of move of all the motors, if needed
        if simultaneous:
            for key in motors_positions_dict:
                mot_hwobj_dict[key].wait_ready(timeout)
        self.update_state()

    def get_value_motors(self, motors_list: list | None = None) -> dict:
        """Get the positions of diffractometer motors. If the motors_list is
            empty, return the positions of all the available motors.

        Args:
            List of motor roles.

        Returns:
            Dictionary {motor_role: position}
        """
        mot_pos_dict = {}

        # use only the available motors
        mot_hwobj_dict = self.get_motors()

        if not motors_list:
            for role, motor in mot_hwobj_dict.items():
                try:
                    mot_pos_dict[role] = float(motor.get_value())
                except TypeError:
                    msg = f"No value for {role}"
                    self.log.warning(msg)
            return mot_pos_dict

        for motor in motors_list:
            try:
                mot_pos_dict[str(motor)] = float(mot_hwobj_dict[motor].get_value())
            except KeyError:
                msg = f"Invalid motor name {motor}"
                self.log.exceptionb(msg)
            except TypeError:
                msg = f"No value for {motor}"
                self.log.warning(msg)
        return mot_pos_dict

    def get_state_motors(self, motors_list: list | None = None) -> dict:
        """Get the state of diffractometer motors. If the motors_list is
            empty, return the state of all the available motors.

        Args:
             List of motor roles.

        Returns:
            Dictionary {motor_role: state}
        """
        mot_state_dict = {}

        # use only the available motors
        mot_hwobj_dict = self.get_motors()

        motors_list = motors_list or mot_hwobj_dict.keys()

        if not motors_list:
            for role, motor in mot_hwobj_dict.items():
                try:
                    mot_state_dict[role] = float(motor.get_state())
                except TypeError:
                    msg = f"No value for {role}"
                    self.log.warning(msg)
            return mot_state_dict

        for motor in motors_list:
            try:
                mot_state_dict[str(motor)] = float(mot_hwobj_dict[motor].get_state())
            except KeyError:
                msg = f"Invalid motor name {motor}"
                self.log.exception(msg)
            except TypeError:
                msg = f"No value for {motor}"
                self.log.warning(msg)
        return mot_state_dict

    # -------- Head Type and Modes --------

    @property
    def get_head_type(self) -> DiffractometerHead:
        """Get the head type
        Returns:
            DiffractometerHead member.
        """
        return self.head_type

    @property
    def in_plate_mode(self) -> bool:
        """Check if the head is a plate."""

        return self.get_head_type == DiffractometerHead.PLATE

    @property
    def in_kappa_mode(self) -> bool:
        """Check if the head is MiniKappa."""

        return self.get_head_type == DiffractometerHead.MINI_KAPPA

    @property
    def in_chip_mode(self) -> bool:
        """Check if there is chip configuration of the head."""
        return (
            self.get_head_type == DiffractometerHead.SSX
            and self.current_constraint == DiffractometerConstraint.STILL
        )

    @property
    def in_injector_mode(self) -> bool:
        """Check if there is injector on the head."""
        return (
            self.get_head_type == DiffractometerHead.SSX
            and self.current_constraint == DiffractometerConstraint.INJECTOR
        )

    @property
    def get_head_enum(self):
        """Get the diffractometer head Enum. Used when no import wished."""
        return DiffractometerHead

    def get_chip_configuration(self) -> Union[GonioHeadConfiguration, None]:
        """Get the chip configuration."""

        data = None

        if Path(self.chip_definition_file).is_file():
            with Path(self.chip_definition_file).open("r") as _f:
                chip_def = json.load(_f)
                try:
                    data = GonioHeadConfiguration(**chip_def)
                except ValidationError:
                    msg = f"Validation error in {self.chip_definition_file}"
                    self.log.exception(msg)
        return data

    def set_head_configuration(self, str_data: str) -> None:
        """Write the chip configuration in the chip configuration json file.
        Args:
            String containing the configuration.
        """
        data = json.loads(str_data)

        if Path(self.chip_definition_file).is_file():
            with Path(self.chip_definition_file).open("w+") as _f:
                try:
                    GonioHeadConfiguration(**data)
                except ValidationError:
                    msg = f"Validation error in {self.chip_definition_file}"
                    self.log.exception(msg)
                else:
                    _f.write(json.dumps(data, indent=4))

    def set_chip_layout(self, layout_name: str) -> bool:
        """Choose the chip configuration layout.
        Args:
            The layout name.
        """
        data = self.get_head_configuration().dict()
        data["current"] = layout_name
        self.set_head_configuration(json.dumps(data))
        return True

    # -------- Phases --------

    def set_phase(self, value: DiffractometerPhase | str, timeout: float | None = None):
        """Sets diffractometer to selected phase.

        Args:
            value: DiffractometerPhase or string value
            timeout: timeout [s],
                     If timeout = 0: return at once and do not wait;
                     if timeout is None: wait forever (default).
        """
        if not isinstance(value, DiffractometerPhase):
            value = self.value_to_enum(value, DiffractometerPhase)
        if value != DiffractometerPhase.UNKNOWN:
            self.update_state(HardwareObjectState.BUSY)
            self._set_phase(value)
            if timeout == 0:
                return
            self.wait_ready(timeout)

    def _set_phase(self, value: DiffractometerPhase):
        """Specific implementation to set the diffractometer to selected phase

        Args:
            value: requested phase.
        """

    def get_phase(self) -> DiffractometerPhase:
        """Get the current phase."""
        return self.current_phase

    def get_phase_list(self) -> list:
        """Return a list of all the defined phases."""
        phase_list = []
        for member in DiffractometerPhase:
            _nam = member.name
            if _nam not in ["IN", "OUT", "UNKNOWN"]:
                phase_list.append(_nam)
        return phase_list

    @property
    def get_phase_enum(self):
        """Get the phase Enum. Used when no import wished."""
        return DiffractometerPhase

    def update_phase(self, value: DiffractometerPhase | None = None):
        """Update the phase value, Emit phaseChanged signal.

        Args:
            value: DiffractometerPhase member. Optional.
        """
        if value is None:
            value = self.get_phase()
        if not isinstance(value, DiffractometerPhase):
            value = DiffractometerPhase(value)
        if self.current_phase != value:
            self.current_phase = value
            self.emit("phaseChanged", (value.name,))

    # -------- Constraints --------

    def set_constraint(
        self, value: DiffractometerConstraint, timeout: float | None = None
    ):
        """Sets diffractometer to selected constraint.

        Args:
            value: DiffractometerConstraint member.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).
        """
        if isinstance(value, DiffractometerConstraint):
            constraint = value
        else:
            constraint = self.value_to_enum(value, DiffractometerConstraint)

        self._set_constraint(constraint)
        self._update_value(constraint, value_cmp=self.get_constraint())
        if timeout == 0:
            return
        self.wait_ready(timeout)

    def _set_constraint(self, value: DiffractometerConstraint):
        """Specific implementation to set the diffractometer to selected
        constraint.
        """

    def get_constraint(self):
        """Get the current constraint
        Returns:
            (Enum): DiffractometerConstraint member.
        """
        return self.current_constraint

    @property
    def get_constraint_enum(self):
        """Get the constraints Enum. Used when no import wished."""
        return DiffractometerConstraint

    # -------- data acquisition scans --------
    def do_oscillation_scan(self, *args, **kwargs):
        """Do an oscillation scan."""
        raise NotImplementedError

    def do_line_scan(self, *args, **kwargs):
        """Do a line (helical) scan."""
        raise NotImplementedError

    def do_mesh_scan(self, *args, **kwargs):
        """Do a mesh scan."""
        raise NotImplementedError

    def do_still_scan(self, *args, **kwargs):
        """Do a zero oscillation acquisition."""
        raise NotImplementedError

    def do_characterisation_scan(self, *args, **kwargs):
        """Do characterisation."""
        raise NotImplementedError

    def _update_value(self, value=None, value_cmp=None):
        """Check if the value has changed. Emits signal valueChanged.

        Args:
            value: value of any type
            value_cmp: Value to compare with.
        """
        curr_value = None
        if value_cmp and value is None:
            curr_value = value_cmp

        if value != curr_value:
            self.emit("valueChanged", (value,))

    # -------- auxilarly methods --------

    def value_to_enum(self, value, which_enum):
        """Tranform a value to Enum

        Args:
           value(str, int, float, tuple, list): value
           which_enum (Enum): The enum to be checked.

        Returns:
            (Enum): Enum member, corresponding to the value or UNKNOWN.
        """
        try:
            return which_enum[value]
        except ValueError:
            for evar in which_enum:
                if isinstance(evar.value, (tuple, list)) and (value in evar.value):
                    return evar
        return which_enum.UNKNOWN



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Example xml file:

.. code-block:: xml

 <object class = "DiffractometerMockup">
  <username>MD2S</username>
  <motors>
    <device role="omega" hwrid="/diff-omega-mockup"/>
    <device role="phiy" hwrid="/diff-phiy-mockup"/>
    <device role="phiz" hwrid="/diff-phiz-mockup"/>
    <device role="phiz" hwrid="/diff-phiz-mockup"/>
    <device role="sampx" hwrid="/diff-sampx-mockup"/>
    <device role="sampy" hwrid="/diff-sampy-mockup"/>
    <device role="kappa" hwrid="/diff-kappa-mockup"/>
    <device role="kappa_phi" hwrid="/diff-kappaphi-mockup"/>
  </motors>
  <nstate_equipment>
    <object role="fshutter" href="/fast-shutter-mockup"/>
    <object role="beamstop" href="/beamstop-mockup"/>
  </nstate_equipment>
 </object>
"""

import random
import time
from warnings import warn

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractDiffractometer import (
    AbstractDiffractometer,
    DiffractometerConstraint,
    DiffractometerHead,
    DiffractometerPhase,
)


class DiffractometerMockup(AbstractDiffractometer):
    """
    Descript. :
    """

    def init(self):
        super().init()
        self.head_type = DiffractometerHead.MINI_KAPPA
        self.current_phase = DiffractometerPhase.CENTRE
        self.current_constraint = DiffractometerConstraint.RELEASE
        self.update_state(HardwareObjectState.READY)
        for mot in self.motors_hwobj_dict.values():
            mot.set_value(random.uniform(0.0, 8.8))
        self.omega.set_value(random.uniform(0, 359.9))

    def abort(self):
        self.update_state(HardwareObjectState.READY)

    def wait_status_ready(self, timeout=None):
        return True

    def save_centring_positions(self):
        """This is a place holder, as needed for some of the diffractometers"""

    def do_oscillation_scan(self, *args, **kwargs):
        if self.in_kappa_mode:
            self.update_state(HardwareObjectState.BUSY)
            time.sleep(random.uniform(0, 2.2))
        self.update_state(HardwareObjectState.READY)

    def get_pixels_per_mm(self):
        """Get the pixel/mm values.
        Returns:
            (tuple): x,y [pixel/mm]
        """
        # calibration values
        x_calib = 0.000444
        y_calib = 0.000446
        return 1.0 / x_calib, 1.0 / y_calib

    def move_motors(self, motors_positions_dict):
        warn(
            "move_motors is deprecated, please use set_value_motors instead",
            DeprecationWarning,
        )
        self.set_value_motors(motors_positions_dict)

    def get_positions(self):
        warn(
            "get_positions is deprecated, please use get_value_motors instead",
            DeprecationWarning,
        )
        return self.get_value_motors()

    def get_current_phase(self):
        warn(
            "get_current_phase is deprecated, please use get get_phase instead",
            DeprecationWarning,
        )
        return self.get_phase().name

    def _set_phase(self, value: DiffractometerPhase):
        """Set a phase."""
        self.current_phase = value
        self.update_state(self.STATES.READY)

    def _set_constraint(self, value):
        self.current_constraint = value



import abc
import collections
import errno
import logging
import os
import socket
import sys
import time

import autoprocessing
import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.TaskUtils import cleanup, error_cleanup, task

BeamlineControl = collections.namedtuple(
    "BeamlineControl",
    [
        "diffractometer",
        "sample_changer",
        "lims",
        "safety_shutter",
        "machine_current",
        "cryo_stream",
        "energy",
        "resolution",
        "detector_distance",
        "transmission",
        "undulators",
        "flux",
        "detector",
        "beam_info",
    ],
)

BeamlineConfig = collections.namedtuple(
    "BeamlineConfig",
    [
        "synchrotron_name",
        "directory_prefix",
        "default_exposure_time",
        "minimum_exposure_time",
        "detector_fileext",
        "detector_type",
        "detector_manufacturer",
        "detector_binning_mode",
        "detector_model",
        "detector_px",
        "detector_py",
        "undulators",
        "focusing_optic",
        "monochromator_type",
        "beam_divergence_vertical",
        "beam_divergence_horizontal",
        "polarisation",
        "maximum_phi_speed",
        "minimum_phi_oscillation",
        "input_files_server",
    ],
)


class AbstractMultiCollect(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.bl_control = BeamlineControl(*[None] * 14)
        self.bl_config = BeamlineConfig(*[None] * 20)
        self.data_collect_task = None
        self.oscillation_task = None
        self.oscillations_history = []
        self.current_lims_sample = None
        self.__safety_shutter_close_task = None
        self.run_without_loop = None
        self.run_autoprocessing = None
        # wait for the 1st image from detector for 30 seconds by default
        self.first_image_timeout = 30

        self.mesh = None
        self.mesh_num_lines = None
        self.mesh_total_nb_frames = None
        self.mesh_range = None
        self.mesh_center = None

        self.number_of_snapshots = 4

    def setControlObjects(self, **control_objects):
        self.bl_control = BeamlineControl(**control_objects)

    def setBeamlineConfiguration(self, **configuration_parameters):
        self.bl_config = BeamlineConfig(**configuration_parameters)

    @abc.abstractmethod
    @task
    def data_collection_hook(self, data_collect_parameters):
        pass

    @task
    def data_collection_end_hook(self, data_collect_parameters):
        pass

    def _bliss_data_collection_hook(self, data_collect_parameters):
        pass

    @abc.abstractmethod
    @task
    def close_fast_shutter(self):
        pass

    @abc.abstractmethod
    @task
    def move_motors(self, motor_position_dict):
        return

    def open_safety_shutter(self):
        pass

    def open_detector_cover(self):
        """Placeholder for implementing opening of the detector cover"""

    def close_detector_cover(self):
        """Placeholder for implementing closing of the detector cover"""

    def safety_shutter_opened(self):
        return False

    def close_safety_shutter(self):
        pass

    def prepare_intensity_monitors(self):
        pass

    @abc.abstractmethod
    def do_oscillation(
        self, start, end, exptime, number_of_images, shutterless, first_frame
    ):
        pass

    @abc.abstractmethod
    def prepare_oscillation(
        self, start, osc_range, exptime, number_of_images, shutterless, first_frame
    ):
        pass

    @abc.abstractmethod
    @task
    def set_detector_filenames(
        self, is_first_frame, frame_number, start, filename, shutterless
    ):
        pass

    @abc.abstractmethod
    def last_image_saved(self):
        pass

    @abc.abstractmethod
    @task
    def prepare_acquisition(
        self, take_dark, start, osc_range, exptime, npass, number_of_images, comment
    ):
        pass

    @abc.abstractmethod
    @task
    def start_acquisition(self, exptime, npass, first_frame, shutterless):
        pass

    @abc.abstractmethod
    @task
    def stop_acquisition(self):
        # detector's readout
        pass

    @abc.abstractmethod
    @task
    def write_image(self, last_frame):
        pass

    @abc.abstractmethod
    @task
    def reset_detector(self):
        pass

    @abc.abstractmethod
    @task
    def data_collection_cleanup(self):
        pass

    @abc.abstractmethod
    def get_wavelength(self):
        pass

    @abc.abstractmethod
    def get_undulators_gaps(self):
        pass

    @abc.abstractmethod
    def get_resolution_at_corner(self):
        pass

    @abc.abstractmethod
    def get_beam_size(self):
        pass

    @abc.abstractmethod
    def get_slit_gaps(self):
        pass

    @abc.abstractmethod
    def get_beam_shape(self):
        pass

    @abc.abstractmethod
    def get_beam_centre(self):
        pass

    @abc.abstractmethod
    def get_machine_current(self):
        pass

    @abc.abstractmethod
    def get_machine_fill_mode(self):
        pass

    @abc.abstractmethod
    def get_machine_message(self):
        pass

    @abc.abstractmethod
    def get_cryo_temperature(self):
        pass

    @abc.abstractmethod
    def store_image_in_lims(self, frame, first_frame, last_frame):
        pass

    @abc.abstractmethod
    def set_fast_characterisation(self, value: bool):
        """
        Use fast characterisation

        Args:
            value: True if to use fast characterisation otherwise False
        """

    @abc.abstractmethod
    @task
    def generate_image_jpeg(self, filename, jpeg_path, jpeg_thumbnail_path):
        pass

    def get_sample_info_from_parameters(self, parameters: dict):
        """Returns sample_id, sample_location and sample_code from data
           collection parameters.
        Args:
            parameters: Dictionary with the data collection parameters
        """
        sample_info = parameters.get("sample_reference")
        try:
            sample_id = int(sample_info.get("blSampleId"))
        except (AttributeError, TypeError):
            sample_id = None

        try:
            sample_code = sample_info["code"]
        except KeyError:
            sample_code = None

        sample_location = None

        try:
            sample_container_number = int(sample_info["container_reference"])
        except KeyError:
            pass
        else:
            try:
                vial_number = int(sample_info["sample_location"])
            except KeyError:
                pass
            else:
                sample_location = (sample_container_number, vial_number)

        return sample_id, sample_location, sample_code

    def create_directories(self, *args):
        for directory in args:
            try:
                os.makedirs(directory)
            except os.error as e:
                if e.errno != errno.EEXIST:
                    raise

    def adxv_notify(self, image_filename: str, image_num: int = 1):
        """
        Notify ADXV of new image

        Args:
           image_filename: full path to image file
           image_num: image number within image file to open (if it contains
                      multiple images i.e HDF5)
        """
        logging.getLogger("HWR").info(f"ADXV notify {image_filename}")
        adxv_host = self.get_property("adxv_host", "localhost")
        adxv_port = int(self.get_property("adxv_port", "8100"))

        try:
            adxv_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            adxv_socket.connect((adxv_host, adxv_port))
            adxv_socket.sendall(
                f"load_image {image_filename}\n slab {image_num}\n".encode()
            )
            adxv_socket.close()
        except Exception:
            logging.getLogger("HWR").exception("")
        else:
            pass

    def take_snapshots(self, dc_params):
        # do nothing if no snapshots asked
        if not self.number_of_snapshots:
            return

        snapshot_directory = dc_params["fileinfo"]["archive_directory"]

        if HWR.beamline.diffractometer.in_plate_mode:
            if self.number_of_snapshots > 0:
                self.number_of_snapshots = 1

        if not os.path.exists(snapshot_directory):
            self.create_directories(snapshot_directory)

        image_path_list = []

        for snapshot_index in range(self.number_of_snapshots):
            snapshot_filename = os.path.join(
                snapshot_directory,
                "%s_%s_%s.snapshot.jpeg"
                % (
                    dc_params["fileinfo"]["prefix"],
                    dc_params["fileinfo"]["run_number"],
                    (snapshot_index + 1),
                ),
            )

            image_path_list.append(snapshot_filename)
            dc_params["xtalSnapshotFullPath%i" % (snapshot_index + 1)] = (
                snapshot_filename
            )

        HWR.beamline.sample_view.take_acq_snapshot(image_path_list)

    @abc.abstractmethod
    def set_helical(self, helical_on):
        pass

    @abc.abstractmethod
    def set_helical_pos(self, helical_pos):
        pass

    def prepare_wedges_to_collect(
        self, start, nframes, osc_range, subwedge_size=1, overlap=0
    ):
        if overlap == 0:
            wedge_sizes_list = [nframes // subwedge_size] * subwedge_size
        else:
            wedge_sizes_list = [subwedge_size] * (nframes // subwedge_size)
        remaining_frames = nframes % subwedge_size
        if remaining_frames:
            wedge_sizes_list.append(remaining_frames)

        wedges_to_collect = []

        for wedge_size in wedge_sizes_list:
            wedges_to_collect.append((start, wedge_size))
            start += wedge_size * osc_range - overlap

        return wedges_to_collect

    def update_oscillations_history(self, data_collect_parameters):
        sample_id, sample_code, sample_location = self.get_sample_info_from_parameters(
            data_collect_parameters
        )
        self.oscillations_history.append(
            (sample_id, sample_code, sample_location, data_collect_parameters)
        )
        return len(self.oscillations_history), sample_id, sample_code, sample_location

    @abc.abstractmethod
    def get_archive_directory(self, directory):
        pass

    @abc.abstractmethod
    def prepare_input_files(
        self, files_directory, prefix, run_number, process_directory
    ):
        """Return XDS input file directory"""

    @abc.abstractmethod
    @task
    def write_input_files(self, collection_id):
        """Write the input files place holder."""

    def execute_collect_without_loop(self, data_collect_parameters):
        return

    def emit_progress(self, progress):
        if progress == 0:
            self.emit("collectReady", (False,))
            self.emit("collectStarted", (None, 1))

            self.emit("collectImageTaken", progress)
        elif progress == 1:
            self.emit(
                "collectEnded",
                None,
                "Data collection successful",
            )
            self.emit("collectReady", (True,))
        elif progress > 0:
            self.emit("collectImageTaken", progress)
        elif progress == -1:
            self.emit(
                "collectOscillationFailed",
                (None, False, "Collection failed", 0, 0),
            )

    def do_collect(self, owner, data_collect_parameters):
        if self.__safety_shutter_close_task is not None:
            self.__safety_shutter_close_task.kill()

        logging.getLogger("user_level_log").info("Closing fast shutter")
        self.close_fast_shutter()

        # reset collection id on each data collect
        self.collection_id = None

        # Preparing directory path for images and processing files
        # creating image file template and jpegs files templates
        file_parameters = data_collect_parameters["fileinfo"]

        file_parameters["suffix"] = self.bl_config.detector_fileext
        image_file_template = (
            "%(prefix)s_%(run_number)s_%%04d.%(suffix)s" % file_parameters
        )
        file_parameters["template"] = image_file_template

        archive_directory = self.get_archive_directory(file_parameters["directory"])
        data_collect_parameters["archive_dir"] = archive_directory

        if archive_directory:
            jpeg_filename = "%s.jpeg" % os.path.splitext(image_file_template)[0]
            thumb_filename = "%s.thumb.jpeg" % os.path.splitext(image_file_template)[0]
            jpeg_file_template = os.path.join(archive_directory, jpeg_filename)
            jpeg_thumbnail_file_template = os.path.join(
                archive_directory, thumb_filename
            )
        else:
            jpeg_file_template = None
            jpeg_thumbnail_file_template = None

        # database filling
        if HWR.beamline.lims:
            data_collect_parameters["collection_start_time"] = time.strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            if HWR.beamline.machine_info is not None:
                logging.getLogger("user_level_log").info(
                    "Getting synchrotron filling mode"
                )
                data_collect_parameters["synchrotronMode"] = (
                    self.get_machine_fill_mode()
                )
            data_collect_parameters["status"] = "failed"

            logging.getLogger("user_level_log").info("Storing data collection in LIMS")
            (self.collection_id, detector_id) = HWR.beamline.lims.store_data_collection(
                data_collect_parameters, self.bl_config
            )

            data_collect_parameters["collection_id"] = self.collection_id

            if detector_id:
                data_collect_parameters["detector_id"] = detector_id

        # Creating the directory for images and processing information
        logging.getLogger("user_level_log").info(
            "Creating directory for images and processing"
        )

        self._bliss_data_collection_hook(data_collect_parameters)

        self.create_directories(
            file_parameters["directory"], file_parameters["process_directory"]
        )
        (
            self.xds_directory,
            self.mosflm_directory,
            self.hkl2000_directory,
        ) = self.prepare_input_files(
            file_parameters["directory"],
            file_parameters["prefix"],
            file_parameters["run_number"],
            file_parameters["process_directory"],
        )
        data_collect_parameters["xds_dir"] = self.xds_directory

        logging.getLogger("user_level_log").info("Getting sample information")
        sample_id, sample_location, sample_code = self.get_sample_info_from_parameters(
            data_collect_parameters
        )
        data_collect_parameters["blSampleId"] = sample_id

        if HWR.beamline.sample_changer is not None:
            try:
                data_collect_parameters["actualSampleBarcode"] = (
                    HWR.beamline.sample_changer.get_loaded_sample().get_id()
                )
                data_collect_parameters["actualContainerBarcode"] = (
                    HWR.beamline.sample_changer.get_loaded_sample()
                    .get_container()
                    .get_id()
                )

                logging.getLogger("user_level_log").info("Getting loaded sample coords")
                (
                    basket,
                    vial,
                ) = HWR.beamline.sample_changer.get_loaded_sample().get_coords()

                data_collect_parameters["actualSampleSlotInContainer"] = vial
                data_collect_parameters["actualContainerSlotInSC"] = basket
            except Exception:
                data_collect_parameters["actualSampleBarcode"] = None
                data_collect_parameters["actualContainerBarcode"] = None
        else:
            data_collect_parameters["actualSampleBarcode"] = None
            data_collect_parameters["actualContainerBarcode"] = None

        centring_info = {}
        try:
            logging.getLogger("user_level_log").info("Getting centring status")
            centring_status = HWR.beamline.sample_view.get_centring_status()
        except Exception:
            logging.getLogger("HWR").exception("")
        else:
            centring_info = dict(centring_status)

        # Save sample centring positions
        positions_str = ""
        motors = centring_info.get(
            "motors", {}
        )  # .update(centring_info.get("extraMotors", {}))
        motors_to_move_before_collect = data_collect_parameters.setdefault("motors", {})

        for motor, pos in motors.items():
            if motor in motors_to_move_before_collect:
                continue
            motors_to_move_before_collect[motor] = pos

        current_diffractometer_position = HWR.beamline.sample_view.get_positions()

        for motor in motors_to_move_before_collect:
            if motors_to_move_before_collect[motor] is not None:
                try:
                    if current_diffractometer_position[motor] is not None:
                        positions_str += "%s=%f " % (
                            motor,
                            current_diffractometer_position[motor],
                        )
                except Exception:
                    logging.getLogger("HWR").exception("")

        positions_str = ""

        for motor, pos in motors_to_move_before_collect.items():
            if pos is not None and motor is not None:
                positions_str += motor + ("=%f" % pos) + " "

        data_collect_parameters["actualCenteringPosition"] = positions_str.strip()

        self.move_motors(motors_to_move_before_collect)
        HWR.beamline.diffractometer.save_centring_positions()

        if data_collect_parameters.get("take_snapshots", False):
            logging.getLogger("user_level_log").info(
                f"Taking sample ({self.number_of_snapshots}) snapshosts"
            )
            self.take_snapshots(data_collect_parameters)
        centring_info = HWR.beamline.sample_view.get_centring_status()
        # move *again* motors, since taking snapshots may change positions
        logging.getLogger("user_level_log").info(
            "Moving motors to centered position: %r", motors_to_move_before_collect
        )
        self.move_motors(motors_to_move_before_collect)

        if HWR.beamline.lims:
            try:
                if self.current_lims_sample:
                    self.current_lims_sample["lastKnownCentringPosition"] = (
                        positions_str
                    )
                    logging.getLogger("user_level_log").info(
                        "Updating sample information in LIMS"
                    )
                    HWR.beamline.lims.update_bl_sample(self.current_lims_sample)
            except Exception:
                logging.getLogger("HWR").exception(
                    "Could not update sample information in LIMS"
                )

        if centring_info.get("images"):
            # Save snapshots
            snapshot_directory = self.get_archive_directory(
                file_parameters["directory"]
            )

            try:
                logging.getLogger("user_level_log").info(
                    "Creating snapshosts directory: %r", snapshot_directory
                )
                self.create_directories(snapshot_directory)
            except Exception:
                logging.getLogger("HWR").exception("Error creating snapshot directory")
            else:
                snapshot_i = 1
                snapshots = []
                for img in centring_info["images"]:
                    img_data = img[1]
                    snapshot_filename = "%s_%s_%s.snapshot.jpeg" % (
                        file_parameters["prefix"],
                        file_parameters["run_number"],
                        snapshot_i,
                    )
                    full_snapshot = os.path.join(snapshot_directory, snapshot_filename)

                    try:
                        f = open(full_snapshot, "w")
                        logging.getLogger("user_level_log").info(
                            "Saving snapshot %d", snapshot_i
                        )
                        f.write(img_data)
                    except Exception:
                        logging.getLogger("HWR").exception("Could not save snapshot!")
                        try:
                            f.close()
                        except Exception:
                            logging.getLogger("HWR").exception("")

                    data_collect_parameters["xtalSnapshotFullPath%i" % snapshot_i] = (
                        full_snapshot
                    )

                    snapshots.append(full_snapshot)
                    snapshot_i += 1

            try:
                data_collect_parameters["centeringMethod"] = centring_info["method"]
            except Exception:
                data_collect_parameters["centeringMethod"] = None

        if HWR.beamline.lims:
            try:
                logging.getLogger("user_level_log").info(
                    "Updating data collection in LIMS"
                )
                if "kappa" in data_collect_parameters["actualCenteringPosition"]:
                    data_collect_parameters["oscillation_sequence"][0]["kappaStart"] = (
                        current_diffractometer_position["kappa"]
                    )
                    data_collect_parameters["oscillation_sequence"][0]["phiStart"] = (
                        current_diffractometer_position["kappa_phi"]
                    )
                HWR.beamline.lims.update_data_collection(data_collect_parameters)
            except Exception:
                logging.getLogger("HWR").exception(
                    "Could not update data collection in LIMS"
                )

        oscillation_parameters = data_collect_parameters["oscillation_sequence"][0]
        sample_id = data_collect_parameters["blSampleId"]
        subwedge_size = oscillation_parameters.get("reference_interval", 1)

        wedges_to_collect = self.prepare_wedges_to_collect(
            oscillation_parameters["start"],
            oscillation_parameters["number_of_images"],
            oscillation_parameters["range"],
            subwedge_size,
            oscillation_parameters["overlap"],
        )
        nframes = sum([wedge_size for _, wedge_size in wedges_to_collect])

        # Added exposure time for ProgressBarBrick.
        # Extra time for each collection needs to be added (in this case 0.04)
        self.emit(
            "collectNumberOfFrames",
            nframes,
            oscillation_parameters["exposure_time"] + 0.04,
        )

        start_image_number = oscillation_parameters["start_image_number"]

        if data_collect_parameters["skip_images"]:
            for _start, wedge_size in wedges_to_collect[:]:
                filename = image_file_template % start_image_number
                file_location = file_parameters["directory"]
                file_path = os.path.join(file_location, filename)
                if os.path.isfile(file_path):
                    logging.info("Skipping existing image %s", file_path)
                    del wedges_to_collect[0]
                    start_image_number += wedge_size
                    nframes -= wedge_size
                else:
                    # images have to be consecutive
                    break

        if nframes == 0:
            return

        # data collection
        self.first_image_timeout = 30 + oscillation_parameters["exposure_time"]
        self.data_collection_hook(data_collect_parameters)

        if "transmission" in data_collect_parameters:
            logging.getLogger("user_level_log").info(
                "Setting transmission to %f", data_collect_parameters["transmission"]
            )
            try:
                HWR.beamline.transmission.set_value(
                    data_collect_parameters["transmission"], timeout=30
                )
            except RuntimeError:
                logging.getLogger("user_level_log").info(
                    "Failed to set transmission to %f",
                    data_collect_parameters["transmission"],
                )
                raise

        if "wavelength" in data_collect_parameters:
            logging.getLogger("user_level_log").info(
                "Setting wavelength to %f", data_collect_parameters["wavelength"]
            )
            try:
                HWR.beamline.energy.set_wavelength(
                    data_collect_parameters["wavelength"], timeout=600
                )
            except RuntimeError:
                logging.getLogger("user_level_log").info(
                    "Failed to set wavelength to %f",
                    data_collect_parameters["wavelength"],
                )
                raise
        elif "energy" in data_collect_parameters:
            logging.getLogger("user_level_log").info(
                "Setting energy to %f", data_collect_parameters["energy"]
            )
            try:
                HWR.beamline.energy.set_value(
                    data_collect_parameters["energy"], timeout=60 * 10
                )
            except RuntimeError:
                logging.getLogger("user_level_log").info(
                    "Failed to set energy to %f", data_collect_parameters["energy"]
                )
                raise

        if "resolution" in data_collect_parameters:
            resolution = data_collect_parameters["resolution"]["upper"]
            logging.getLogger("user_level_log").info(
                "Setting resolution to %f", resolution
            )
            try:
                self.open_detector_cover()
                HWR.beamline.resolution.set_value(resolution, timeout=3500)
            except (AttributeError, RuntimeError):
                logging.getLogger("user_level_log").info(
                    "Failed to set resolution to %f", resolution
                )
                raise
        elif "detector_distance" in oscillation_parameters:
            logging.getLogger("user_level_log").info(
                "Moving detector to %f", data_collect_parameters["detector_distance"]
            )
            try:
                HWR.beamline.detector.distance.set_value(
                    oscillation_parameters["detector_distance"], timeout=3500
                )
            except RuntimeError:
                logging.getLogger("user_level_log").info(
                    "Failed to set detector distance to %f",
                    data_collect_parameters["detector_distance"],
                )
                raise

        with cleanup(self.data_collection_cleanup):
            # if not self.safety_shutter_opened():
            self.open_safety_shutter()

            try:
                HWR.beamline.beam.wait_for_beam()
                HWR.beamline.cryo.wait_temperature()
            except AttributeError:
                logging.getLogger("HWR").exception("")

            logging.getLogger("user_level_log").info("Preparing intensity monitors")
            self.prepare_intensity_monitors()

            frame = start_image_number
            osc_range = oscillation_parameters["range"]
            exptime = oscillation_parameters["exposure_time"]
            npass = oscillation_parameters["number_of_passes"]

            # update LIMS
            if HWR.beamline.lims:
                try:
                    logging.getLogger("user_level_log").info(
                        "Gathering data for LIMS update"
                    )
                    data_collect_parameters["flux"] = HWR.beamline.flux.get_value()
                    data_collect_parameters["flux_end"] = data_collect_parameters[
                        "flux"
                    ]
                    data_collect_parameters["wavelength"] = (
                        HWR.beamline.energy.get_wavelength()
                    )
                    data_collect_parameters["detectorDistance"] = (
                        HWR.beamline.detector.distance.get_value()
                    )
                    data_collect_parameters["resolution"] = (
                        HWR.beamline.resolution.get_value()
                    )
                    data_collect_parameters["transmission"] = (
                        HWR.beamline.transmission.get_value()
                    )
                    beam_centre_x, beam_centre_y = self.get_beam_centre()
                    data_collect_parameters["xBeam"] = beam_centre_x
                    data_collect_parameters["yBeam"] = beam_centre_y

                    und = self.get_undulators_gaps()
                    i = 1
                    for jj in self.bl_config.undulators:
                        key = jj.type
                        if key in und:
                            data_collect_parameters["undulatorGap%d" % (i)] = und[key]
                            i += 1
                    data_collect_parameters["resolutionAtCorner"] = (
                        self.get_resolution_at_corner()
                    )
                    beam_size_x, beam_size_y = self.get_beam_size()
                    data_collect_parameters["beamSizeAtSampleX"] = beam_size_x
                    data_collect_parameters["beamSizeAtSampleY"] = beam_size_y
                    data_collect_parameters["beamShape"] = self.get_beam_shape()
                    hor_gap, vert_gap = self.get_slit_gaps()
                    data_collect_parameters["slitGapHorizontal"] = hor_gap
                    data_collect_parameters["slitGapVertical"] = vert_gap

                    logging.getLogger("user_level_log").info(
                        "Updating data collection in LIMS"
                    )
                    HWR.beamline.lims.update_data_collection(data_collect_parameters)
                    logging.getLogger("user_level_log").info(
                        "Done updating data collection in LIMS"
                    )
                except Exception:
                    logging.getLogger("HWR").exception(
                        "Could not store data collection into LIMS"
                    )

            if HWR.beamline.lims and self.bl_config.input_files_server:
                logging.getLogger("user_level_log").info(
                    "Asking for input files writing"
                )
                self.write_input_files(self.collection_id, wait=False)

            # at this point input files should have been written
            # TO DO: agree what parameters will be sent to this function
            if data_collect_parameters.get("processing", False) == "True":
                self.trigger_auto_processing(
                    "before",
                    self.xds_directory,
                    data_collect_parameters["EDNA_files_dir"],
                    data_collect_parameters["anomalous"],
                    data_collect_parameters["residues"],
                    data_collect_parameters["do_inducedraddam"],
                    data_collect_parameters.get("sample_reference", {}).get(
                        "spacegroup", ""
                    ),
                    data_collect_parameters.get("sample_reference", {}).get("cell", ""),
                )
            if self.run_without_loop:
                self.execute_collect_without_loop(data_collect_parameters)
            else:
                for start, wedge_size in wedges_to_collect:
                    logging.getLogger("user_level_log").info(
                        "Preparing acquisition, start=%f, wedge size=%d",
                        start,
                        wedge_size,
                    )

                    self.prepare_acquisition(
                        1 if data_collect_parameters.get("dark", 0) else 0,
                        start,
                        osc_range,
                        exptime,
                        npass,
                        wedge_size,
                        data_collect_parameters.get("comment", ""),
                    )
                    data_collect_parameters["dark"] = 0

                    i = 0
                    j = wedge_size

                    _total_time_spent = 0
                    _total_exptime = exptime
                    while j > 0:
                        _time_start = time.time()
                        frame_start = start + i * osc_range
                        i += 1

                        filename = image_file_template % frame
                        try:
                            jpeg_full_path = jpeg_file_template % frame
                            jpeg_thumbnail_full_path = (
                                jpeg_thumbnail_file_template % frame
                            )
                        except Exception:
                            jpeg_full_path = None
                            jpeg_thumbnail_full_path = None
                        file_location = file_parameters["directory"]
                        file_path = os.path.join(file_location, filename)

                        self.set_detector_filenames(
                            i == 1,
                            frame,
                            frame_start,
                            str(file_path),
                            data_collect_parameters.get("shutterless", True),
                        )

                        osc_start, osc_end = self.prepare_oscillation(
                            frame_start,
                            osc_range,
                            exptime,
                            wedge_size,
                            data_collect_parameters.get("shutterless", True),
                            j == wedge_size,
                        )

                        with error_cleanup(self.reset_detector):
                            self.start_acquisition(
                                exptime,
                                npass,
                                j == wedge_size,
                                data_collect_parameters.get("shutterless", True),
                            )
                            self.do_oscillation(
                                osc_start,
                                osc_end,
                                exptime,
                                wedge_size,
                                data_collect_parameters.get("shutterless", True),
                                j == wedge_size,
                            )

                            self.write_image(j == 1)

                        # Store image in lims
                        if HWR.beamline.lims:
                            if self.store_image_in_lims(frame, j == wedge_size, j == 1):
                                lims_image = {
                                    "dataCollectionId": self.collection_id,
                                    "fileName": filename,
                                    "fileLocation": file_location,
                                    "imageNumber": frame,
                                    "measuredIntensity": HWR.beamline.flux.get_value(),
                                    "synchrotronCurrent": self.get_machine_current(),
                                    "machineMessage": self.get_machine_message(),
                                    "temperature": self.get_cryo_temperature(),
                                }

                                if archive_directory:
                                    lims_image["jpegFileFullPath"] = jpeg_full_path
                                    lims_image["jpegThumbnailFileFullPath"] = (
                                        jpeg_thumbnail_full_path
                                    )

                                try:
                                    HWR.beamline.lims.store_image(lims_image)
                                except Exception:
                                    logging.getLogger("HWR").exception(
                                        "Could not store store image in LIMS"
                                    )

                                self.generate_image_jpeg(
                                    str(file_path),
                                    str(jpeg_full_path),
                                    str(jpeg_thumbnail_full_path),
                                    wait=False,
                                )

                        if data_collect_parameters.get("processing", False) == "True":
                            self.trigger_auto_processing(
                                "image",
                                self.xds_directory,
                                data_collect_parameters["EDNA_files_dir"],
                                data_collect_parameters["anomalous"],
                                data_collect_parameters["residues"],
                                data_collect_parameters["do_inducedraddam"],
                                data_collect_parameters.get("sample_reference", {}).get(
                                    "spacegroup", ""
                                ),
                                data_collect_parameters.get("sample_reference", {}).get(
                                    "cell", ""
                                ),
                            )

                        last_image_saved = self.last_image_saved(
                            _total_exptime, exptime, wedge_size
                        )

                        if data_collect_parameters.get("shutterless"):
                            msg = "Timeout waiting for detector trigger, no image taken"
                            with gevent.Timeout(
                                self.first_image_timeout,
                                RuntimeError(msg),
                            ):
                                if last_image_saved <= 0:
                                    last_image_saved = self.last_image_saved(
                                        _total_exptime, exptime, wedge_size
                                    )

                            last_image_saved = self.last_image_saved(
                                _total_exptime, exptime, wedge_size
                            )

                            if last_image_saved < wedge_size:
                                time.sleep(exptime)
                                last_image_saved = self.last_image_saved(
                                    _total_exptime, exptime, wedge_size
                                )
                            frame = max(
                                start_image_number + 1,
                                start_image_number + last_image_saved - 1,
                                frame + 1,
                            )
                            self.emit("collectImageTaken", frame)
                            j = wedge_size - last_image_saved
                        else:
                            j -= 1
                            self.emit("collectImageTaken", frame)
                            frame += 1
                            if j == 0:
                                break

                        _total_time_spent += time.time() - _time_start
                        _total_exptime += exptime

            # Bug fix for MD2/3(UP): diffractometer still has things to do
            # even after the last frame is taken (decelerate motors and
            # possibly download diagnostics) so we cannot trigger the cleanup
            # (that will send an abort on the diffractometer) as soon as
            # the last frame is counted
            self.diffractometer().wait_status_ready(1000)

        # data collection done
        self.data_collection_end_hook(data_collect_parameters)

    @task
    def loop(self, owner, data_collect_parameters_list):
        failed_msg = "Data collection failed!"
        failed = True
        collections_analyse_params = []

        try:
            self.emit("collectReady", (False,))
            self.emit("collectStarted", (owner, 1))
            for data_collect_parameters in data_collect_parameters_list:
                logging.debug("collect parameters = %r", data_collect_parameters)
                failed = False
                try:
                    # emit signals to make bricks happy
                    (
                        osc_id,
                        sample_id,
                        sample_code,
                        sample_location,
                    ) = self.update_oscillations_history(data_collect_parameters)
                    self.emit(
                        "collectOscillationStarted",
                        (
                            owner,
                            sample_id,
                            sample_code,
                            sample_location,
                            data_collect_parameters,
                            osc_id,
                        ),
                    )
                    data_collect_parameters["status"] = "Running"

                    # now really start collect sequence
                    self.do_collect(owner, data_collect_parameters)
                except Exception:
                    failed = True
                    exc_type, exc_value, exc_tb = sys.exc_info()
                    msg = f"Data collection failed: {exc_value}"
                    logging.exception("Data collection failed")
                    logging.getLogger("user_level_log").info(msg)
                    data_collect_parameters["status"] = (
                        "Data collection failed!"  # Message to be stored in LIMS
                    )
                    failed_msg = "Data collection failed!\n%s" % exc_value
                    self.emit(
                        "collectOscillationFailed",
                        (owner, False, failed_msg, self.collection_id, osc_id),
                    )
                else:
                    data_collect_parameters["status"] = "Data collection successful"

                try:
                    if data_collect_parameters.get("processing", False) == "True":
                        self.trigger_auto_processing(
                            "after",
                            self.xds_directory,
                            data_collect_parameters["EDNA_files_dir"],
                            data_collect_parameters["anomalous"],
                            data_collect_parameters["residues"],
                            data_collect_parameters["do_inducedraddam"],
                            data_collect_parameters.get("sample_reference", {}).get(
                                "spacegroup", ""
                            ),
                            data_collect_parameters.get("sample_reference", {}).get(
                                "cell", ""
                            ),
                        )
                except Exception:
                    logging.getLogger("HWR").exception("")
                else:
                    collections_analyse_params.append(
                        (
                            self.collection_id,
                            self.xds_directory,
                            data_collect_parameters["EDNA_files_dir"],
                            data_collect_parameters["anomalous"],
                            data_collect_parameters["residues"],
                            "reference_interval"
                            in data_collect_parameters["oscillation_sequence"][0],
                            data_collect_parameters["do_inducedraddam"],
                        )
                    )

                if HWR.beamline.lims:
                    data_collect_parameters["flux_end"] = HWR.beamline.flux.get_value()
                    try:
                        HWR.beamline.lims.finalize_data_collection(
                            data_collect_parameters
                        )
                    except Exception:
                        logging.getLogger("HWR").exception(
                            "Could not store data collection into LIMS"
                        )

                if failed:
                    # if one dc fails, stop the whole loop
                    break
                self.emit(
                    "collectOscillationFinished",
                    (
                        owner,
                        True,
                        data_collect_parameters["status"],
                        self.collection_id,
                        osc_id,
                        data_collect_parameters,
                    ),
                )

            if self.get_property("close_safety_shutter_if_idle", False):
                try:
                    self.__safety_shutter_close_task = gevent.spawn_later(
                        10 * 60, self.close_safety_shutter, timeout=10
                    )
                except Exception:
                    logging.exception("Could not close safety shutter")
        finally:
            self.emit(
                "collectEnded",
                owner,
                not failed,
                failed_msg if failed else "Data collection successful",
            )
            self.emit("collectReady", (True,))

    def collect_without_loop(self, data_collect_parameters_list):
        self.emit("collectReady", (False,))
        total_frames = 0
        total_time_sec = 0
        # in principle data_collect_parameters_list always has one element
        # Two options: rename this module to AbstractCollect
        # or fully implement multicollect
        for data_collect_parameters in data_collect_parameters_list:
            total_frames = (
                total_frames
                + data_collect_parameters["oscillation_sequence"][0]["number_of_images"]
            )
            total_time_sec = (
                total_time_sec
                + total_frames
                + 2
                * (
                    data_collect_parameters["oscillation_sequence"][0]["exposure_time"]
                    + 0.02
                )
            )
        self.emit("collectOverallFramesTime", (total_frames, int(total_time_sec)))
        for data_collect_parameters in data_collect_parameters_list:
            self.actual_data_collect_parameters = data_collect_parameters
            (
                self.osc_id,
                sample_id,
                sample_code,
                sample_location,
            ) = self.update_oscillations_history(self.actual_data_collect_parameters)
            self.emit(
                "collectOscillationStarted",
                (
                    self.owner,
                    sample_id,
                    sample_code,
                    sample_location,
                    self.actual_data_collect_parameters,
                    self.osc_id,
                ),
            )
            self.actual_data_collect_parameters["status"] = "Running"
            self.do_collect(self.owner, self.actual_data_collect_parameters)

    def collect(
        self, owner, data_collect_parameters_list
    ):  # , finished_callback=None):
        if self.run_without_loop:
            self.owner = owner
            self.ready_event.clear()
            self.data_collect_task = gevent.spawn(
                self.collect_without_loop, data_collect_parameters_list
            )
            self.ready_event.wait()
            self.ready_event.clear()
        else:
            self.data_collect_task = self.loop(
                owner, data_collect_parameters_list, wait=False
            )
        return self.data_collect_task

    def stop_collect(self, owner=None):
        if self.data_collect_task is not None:
            self.data_collect_task.kill(block=True, timeout=120)

        self.data_collection_cleanup()

    """
    processDataScripts
        Description    : executes a script after the data collection has finished
        Type           : method
    """

    def trigger_auto_processing(
        self,
        process_event,
        xds_dir,
        EDNA_files_dir=None,
        anomalous=None,
        residues=200,
        do_inducedraddam=False,
        spacegroup=None,
        cell=None,
    ):
        # quick fix for anomalous, do_inducedraddam... passed as a string!!!
        # (comes from the queue)
        if isinstance(anomalous, str):
            anomalous = anomalous == "True"
        if isinstance(do_inducedraddam, str):
            do_inducedraddam = do_inducedraddam == "True"
        if isinstance(residues, str):
            try:
                residues = int(residues)
            except ValueError:
                residues = 200

        # residues = zero should be interpreted as if no value was provided
        # use default of 200
        residues = residues or 200

        processAnalyseParams = {}
        processAnalyseParams["EDNA_files_dir"] = EDNA_files_dir

        try:
            if isinstance(xds_dir, list):
                processAnalyseParams["collections_params"] = xds_dir
            else:
                processAnalyseParams["datacollect_id"] = self.collection_id
                processAnalyseParams["xds_dir"] = xds_dir
            processAnalyseParams["anomalous"] = anomalous
            processAnalyseParams["residues"] = residues
            processAnalyseParams["spacegroup"] = spacegroup
            processAnalyseParams["cell"] = cell
        except Exception as err:
            msg = f"DataCollect:processing: {err}"
            logging.getLogger().exception(msg)
        else:
            programs = self.get_property("auto_processing")
            try:
                autoprocessing.start(programs, process_event, processAnalyseParams)
            except Exception:
                logging.getLogger().exception("Error starting processing")

            if process_event == "after" and do_inducedraddam:
                try:
                    autoprocessing.startInducedRadDam(processAnalyseParams)
                except Exception:
                    logging.exception("Error starting induced rad.dam")

    def set_run_autoprocessing(self, status):
        pass

    # specifies the next scan will be a mesh scan
    def set_mesh(self, mesh_on):
        self.mesh = mesh_on

    def set_mesh_scan_parameters(
        self, num_lines, total_nb_frames, mesh_center_param, mesh_range_param
    ):
        """
        sets the mesh scan parameters :
         - vertcal range
         - horizontal range
         - nb lines
         - nb frames per line
         - invert direction (boolean)  # NOT YET DONE
        """
        self.mesh_num_lines = num_lines
        self.mesh_total_nb_frames = total_nb_frames
        self.mesh_range = mesh_range_param
        self.mesh_center = mesh_center_param



"""Class for cameras connected to Lima Tango Device Servers

Example configuration:
----------------------
<object class="TangoLimaVideo">
  <username>Prosilica 1350C</username>
  <tangoname>id23/limaccd/minidiff2</tangoname>
  <bpmname>id23/limabeamviewer/minidiff2</bpmname>
  <interval>15</interval>
  <video_mode>RGB24</video_mode>
</object>

If video mode is not specified, BAYER_RG16 is used by default.
"""

import io
import struct
import time

import gevent
import PyTango
from PIL import Image
from PyTango.gevent import DeviceProxy

from mxcubecore import BaseHardwareObjects


def poll_image(lima_tango_device, video_mode, FORMATS):
    img_data = lima_tango_device.video_last_image

    hfmt = ">IHHqiiHHHH"
    hsize = struct.calcsize(hfmt)
    _, _, _, _, width, height, _, _, _, _ = struct.unpack(hfmt, img_data[1][:hsize])

    raw_data = img_data[1][hsize:]
    _from, _to = FORMATS.get(video_mode, (None, None))

    if _from and _to:
        img = Image.frombuffer(_from, (height, width), raw_data, "raw", _from, 0, 1)

        img_bytes = io.BytesIO()
        img.save(img_bytes, format=_to)
        img = img.tobytes()
    else:
        img = raw_data

    return img, width, height


class TangoLimaVideo(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        super().__init__(name)
        self.__polling = None
        self._video_mode = None
        self._last_image = (0, 0, 0)

        # Dictionary containing conversion information for a given
        # video_mode. The camera video mode is the key and the first
        # index of the tuple contains the corresponding PIL mapping
        # and the second the desried output format. The image is
        # passed on as it is of the video mode is not in the dictionary
        self._FORMATS = {
            "RGB8": ("L", "BMP"),
            "RGB24": ("RGB", "BMP"),
            "RGB32": ("RGBA", "BMP"),
            "NO_CONVERSION": (None, None),
        }

    def init(self):
        self.device = None

        try:
            self._video_mode = self.get_property("video_mode", "RGB24")
            self.device = DeviceProxy(self.get_property("tangoname"))
            # try a first call to get an exception if the device
            # is not exported
            self.device.ping()
        except PyTango.DevFailed as traceback:
            last_error = traceback[-1]
            self.log.error("%s: %s", str(self.name()), last_error.desc)

            self.device = BaseHardwareObjects.Null()
        else:
            if self.get_property("exposure_time"):
                self._sleep_time = float(self.get_property("exposure_time"))
            elif self.get_property("interval"):
                self._sleep_time = float(self.get_property("interval")) / 1000.0

            if self.get_property("control_video", "True"):
                self.log.info("MXCuBE controlling video")

                if self.device.video_live:
                    self.device.video_live = False

                self.device.video_mode = self._video_mode
                self.set_exposure(self._sleep_time)

                self.device.video_live = True
            else:
                self.log.info("MXCuBE NOT controlling video")

        self.update_state(BaseHardwareObjects.HardwareObjectState.READY)

    def get_last_image(self):
        return poll_image(self.device, self._video_mode, self._FORMATS)

    def _do_polling(self, sleep_time):
        lima_tango_device = self.device

        while True:
            data, width, height = poll_image(
                lima_tango_device, self._video_mode, self._FORMATS
            )

            self._last_image = data, width, height
            self.emit("imageReceived", data, width, height, False)
            time.sleep(sleep_time)

    def connect_notify(self, signal):
        if signal == "imageReceived":
            if self.__polling is None:
                self.__polling = gevent.spawn(
                    self._do_polling, self.device.video_exposure
                )

    def get_width(self):
        return self.device.image_width

    def get_height(self):
        return self.device.image_height

    def set_live(self, mode):
        curr_state = self.device.video_live

        if mode:
            if not curr_state:
                self.device.video_live = True
        else:
            if curr_state:
                self.device.video_live = False

    def set_exposure(self, exposure):
        self.device.video_exposure = exposure



from mxcubecore.BaseHardwareObjects import HardwareObject


class PX1Attenuator(HardwareObject):
    stateAttenuator = {
        "ALARM": "error",
        "OFF": "error",
        "RUNNING": "moving",
        "MOVING": "moving",
        "STANDBY": "ready",
        "UNKNOWN": "changed",
        "EXTRACT": "extract",
        "INSERT": "insert",
    }

    def init(self):
        self.state_chan = self.get_channel_object("state")
        self.factor_chan = self.get_channel_object("parser")

        if self.state_chan is not None:
            self.state_chan.connect_signal("update", self.state_changed)

        if self.factor_chan is not None:
            self.factor_chan.connect_signal("update", self.factor_changed)

        self.connected()

    def connected(self):
        self.set_is_ready(True)

    def disconnected(self):
        self.set_is_ready(False)

    def getAttState(self, value=None):
        if not self.state_chan:
            return

        if value is None:
            value = self.state_chan.get_value()

        try:
            state_str = str(value)
            retval = self.stateAttenuator[state_str]
        except Exception:
            value = None

        return retval

    def get_value(self):
        try:
            value = round(float(self.factor_chan.get_value()), 1)
        except Exception:
            value = None

        return value

    def state_changed(self, value=None):
        state_value = self.getAttState(value)
        self.emit("attStateChanged", (state_value,))

    def factor_changed(self, channelValue):
        try:
            value = self.get_value()
        except Exception:
            self.log.error(
                "%s attFactorChanged : received value on channel is not a float value",
                str(self.name()),
            )
        else:
            self.emit("attFactorChanged", (value,))

    def _set_value(self, value):
        self.factor_chan.set_value(value)


def test_hwo(self):
    pass



import logging
import math

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject

"""
Example xml file:
  - for tunable wavelength beamline:
<object class="Energy">
  <object href="/energy" role="energy"/>
  <object href="/bliss" role="controller"/>
  <tunable_energy>True</tunable_energy>
</object>
The energy should have methods get_value, get_limits and move.
If used, the controller should have method moveEnergy.

  - for fixed wavelength beamline:
<object class="Energy">
  <default_energy>12.8123</tunable_energy>
</object>
"""


class Energy(HardwareObject):
    def init(self):
        self.ready_event = gevent.event.Event()
        self.energy_motor = None
        self.tunable = False
        self.moving = None
        self.default_en = None
        self.ctrl = None
        self.en_lims = []

        try:
            self.energy_motor = self.get_object_by_role("energy")
        except KeyError:
            self.log.warning("Energy: error initializing energy motor")

        try:
            self.default_en = self.get_property("default_energy")
        except KeyError:
            self.log.warning("Energy: no default energy")

        try:
            self.tunable = self.get_property("tunable_energy")
        except KeyError:
            self.log.warning("Energy: will set to fixed energy")

        try:
            self.ctrl = self.get_object_by_role("controller")
        except KeyError:
            self.log.info("No controller used")

        if self.energy_motor is not None:
            self.energy_motor.connect("valueChanged", self.energyPositionChanged)
            self.energy_motor.connect("stateChanged", self.energyStateChanged)

    def is_connected(self):
        return True

    def get_value(self):
        if self.energy_motor is not None:
            try:
                return self.energy_motor.get_value()
            except Exception:
                self.log.exception("EnergyHO: could not read current energy")
                return None
        return self.default_en

    def get_wavelength(self):
        current_en = self.get_current_energy()
        if current_en:
            return 12.3984 / current_en
        return None

    def get_limits(self):
        self.log.debug("Get energy limits")
        if not self.tunable:
            energy = self.get_value()
            return (energy, energy)

        if self.energy_motor is not None:
            try:
                self.en_lims = self.energy_motor.get_limits()
                return self.en_lims
            except Exception:
                self.log.exception("EnergyHO: could not read energy motor limits")
                return None
        return None

    def get_wavelength_limits(self):
        self.log.debug("Get wavelength limits")
        if not self.tunable:
            return None
        self.en_lims = self.get_limits()
        if self.en_lims:
            lims = (12.3984 / self.en_lims[1], 12.3984 / self.en_lims[0])
        return lims

    # def start_move_energy(self, value, wait=True):
    #     if not self.tunable:
    #         return False
    #
    #     try:
    #         value = float(value)
    #     except (TypeError, ValueError) as diag:
    #         logging.getLogger("user_level_log").error(
    #             "Energy: invalid energy (%s)" % value
    #         )
    #         return False
    #
    #     current_en = self.get_current_energy()
    #     if current_en:
    #         if math.fabs(value - current_en) < 0.001:
    #             self.moveEnergyCmdFinished(True)
    #     if self.checkLimits(value) is False:
    #         return False
    #
    #     self.moveEnergyCmdStarted()
    #
    #     def change_egy():
    #         try:
    #             self.set_value(value, wait=True)
    #         except Exception:
    #             sys.excepthook(*sys.exc_info())
    #             self.moveEnergyCmdFailed()
    #         else:
    #             self.moveEnergyCmdFinished(True)
    #
    #     if wait:
    #         change_egy()
    #     else:
    #         gevent.spawn(change_egy)

    def moveEnergyCmdStarted(self):
        self.moving = True
        self.emit("moveEnergyStarted", ())

    def moveEnergyCmdFailed(self):
        self.moving = False
        self.emit("moveEnergyFailed", ())

    def moveEnergyCmdAborted(self):
        pass

    def moveEnergyCmdFinished(self, result):
        self.moving = False
        self.emit("moveEnergyFinished", ())

    def checkLimits(self, value):
        self.log.debug("Checking the move limits")
        if self.get_limits():
            if value >= self.en_lims[0] and value <= self.en_lims[1]:
                self.log.info("Limits ok")
                return True
            logging.getLogger("user_level_log").info("Requested value is out of limits")
        return False

    # def start_move_wavelength(self, value, wait=True):
    #     self.log.info("Moving wavelength to (%s)" % value)
    #     return self.startMoveEnergy(12.3984 / value, wait)

    def cancelMoveEnergy(self):
        logging.getLogger("user_level_log").info("Cancel move")
        self.moveEnergy.abort()

    def set_value(self, energy, wait=True):
        current_en = self.get_value()
        pos = math.fabs(current_en - energy)
        if pos < 0.001:
            logging.getLogger("user_level_log").info(
                "Energy: already at %g, not moving", energy
            )
        else:
            logging.getLogger("user_level_log").info(
                "Energy: moving energy to %g", energy
            )
            if pos > 0.02:
                try:
                    if self.ctrl:
                        self.ctrl.change_energy(energy)
                    else:
                        self.execute_command("moveEnergy", energy, wait=True)
                except RuntimeError as AttributeError:
                    self.energy_motor.set_value(energy)
            else:
                self.energy_motor.set_value(energy)

    def energyPositionChanged(self, pos):
        wl = 12.3984 / pos
        if wl:
            self.emit("energyChanged", (pos, wl))
            self.emit("valueChanged", (pos,))

    def energyStateChanged(self, state):
        self.emit("stateChanged", (state,))



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class AdvancedConnectorQueueEntry(BaseQueueEntry):
    """Controls different steps"""

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):
        BaseQueueEntry.__init__(self, view, data_model, view_set_queue_entry)
        self.first_qe = None
        self.second_qe = None

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)

    def execute(self):
        BaseQueueEntry.execute(self)
        first_qe_data_model = self.first_qe.get_data_model()

        if first_qe_data_model.run_online_processing == "XrayCentering":
            best_positions = first_qe_data_model.online_processing_results[
                "aligned"
            ].get("best_positions", [])

            if len(best_positions) > 0:
                best_cpos = best_positions[0]["cpos"]
                helical_model = self.second_qe.get_data_model()

                # logging.getLogger("user_level_log").info(\
                #    "Moving to the best position")
                # HWR.beamline.diffractometer.move_motors(best_cpos)
                # gevent.sleep(2)

                logging.getLogger("user_level_log").info("Rotating 90 degrees")
                HWR.beamline.diffractometer.move_omega_relative(90)
                logging.getLogger("user_level_log").info("Creating a helical line")

                gevent.sleep(2)
                (
                    auto_line,
                    cpos_one,
                    cpos_two,
                ) = HWR.beamline.sample_view.create_auto_line()
                helical_model.acquisitions[
                    0
                ].acquisition_parameters.osc_start = cpos_one.phi
                helical_model.acquisitions[
                    0
                ].acquisition_parameters.centred_position = cpos_one
                helical_model.acquisitions[
                    1
                ].acquisition_parameters.centred_position = cpos_two

                self.second_qe.set_enabled(True)
            else:
                logging.getLogger("user_level_log").warning(
                    "No diffraction found. Cancelling Xray centering"
                )
                self.second_qe.set_enabled(False)



import time
from typing import Optional

import gevent
import numpy as np

from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator


class EPICSActuator(AbstractActuator):
    ACTUATOR_VAL = "val"  # setpoint
    ACTUATOR_RBV = "rbv"  # readback

    def __init__(self, name):
        super().__init__(name)
        self.setpoint = None
        self._nominal_limits = (-1e4, 1e4)
        self.default_timeout = 180
        if not self.unit:
            self.unit = 10**-3

    def init(self):
        super().init()
        self.update_state(self.STATES.READY)
        self.connect(self.get_channel_object("rbv"), "update", self.update_value)

    def hasnt_arrived(self, setpoint):
        readback = self.get_value()
        if not readback:
            return False
        return not np.isclose(readback, setpoint, rtol=self.unit, atol=self.unit)

    def wait_ready(self, timeout: Optional[float] = None):
        self._ready_event.clear()
        is_set = self._ready_event.is_set()
        try:
            with gevent.Timeout(timeout, exception=TimeoutError):
                while not is_set:
                    is_set = self._ready_event.is_set()
                    if not self.hasnt_arrived(self.setpoint):
                        self._ready_event.set()
                    time.sleep(0.15)
        except TimeoutError:
            pvname = self.get_channel_object("").command.pv_name
            self.print_log(
                level="error",
                msg=f"{pvname} motion has timed out.",
            )
        self.update_state(self.STATES.READY)

    def get_value(self):
        return self.get_channel_value(self.ACTUATOR_RBV)

    def _set_value(self, value):
        self.setpoint = value
        self.update_state(self.STATES.BUSY)
        self.set_channel_value(self.ACTUATOR_VAL, value)

    def set_value(self, value, timeout: float = 0):
        if not timeout:
            timeout = self.default_timeout
        try:
            super().set_value(value, timeout)
        except ValueError:
            pvname = self.get_channel_object("rbv").command.pv_name
            msg = f"{pvname} value {value} is outside of actuator limits"
            msg += f": {self._nominal_limits}"
            self.print_log(
                level="error",
                msg=msg,
            )

    def abort(self):
        if self._wait_task is not None:
            self._wait_task.set()
        self.update_state(self.STATES.READY)



"""SpecMotorWSpecPositions

template:
  <object class = "SpecMotorWSpecPositions">
    <username>friendly label for user</username>
    <specname>motor mnemonic in Spec</specname>
    <specversion>host:version</specversion>
    <channel type="spec" name="positions">global associative array in Spec with positions defined</channel>
    <command type="spec" name="setNewPosition">macro in Spec to set new positions from an associative array</command>
    <!-- <GUIstep>default step when moving motor from the GUI</GUIstep>
         <delta>tolerance allowed between real motor position and defined position</delta> -->
  </object>
"""

from mxcubecore.HardwareObjects import SpecMotor


class SpecMotorWSpecPositions(SpecMotor.SpecMotor):
    def __init__(self, *args):
        SpecMotor.SpecMotor.__init__(self, *args)

        self.predefinedPositions = {}
        self.predefinedPositionsNamesList = []

    def init(self):
        chanPositionsArray = self.get_channel_object("positions")
        chanPositionsArray.connect_signal("update", self.positionsArrayChanged)
        self.delta = self.get_property("delta") or 0

    def disconnected(self):
        self.predefinedPositions = {}
        self.predefinedPositionsNamesList = []
        self.emit("newPredefinedPositions", (self.predefinedPositionsNamesList,))

    def positionsArrayChanged(self, positionsArray):
        self.predefinedPositions = positionsArray

        for pos in self.predefinedPositions:
            self.predefinedPositions[pos] = float(self.predefinedPositions[pos])

        self.sortPredefinedPositionsList()
        self.emit("newPredefinedPositions", (self.predefinedPositionsNamesList,))

    def sortPredefinedPositionsList(self):
        self.predefinedPositionsNamesList = list(self.predefinedPositions.keys())
        self.predefinedPositionsNamesList.sort(
            lambda x, y: int(
                round(self.predefinedPositions[x] - self.predefinedPositions[y])
            )
        )

    def motor_positions_changed(self, channelValue):
        SpecMotor.SpecMotor.motor_positions_changed.__func__(self, channelValue)

        pos = float(channelValue)

        for positionName in self.predefinedPositions:
            if (
                self.predefinedPositions[positionName] >= pos - self.delta
                and self.predefinedPositions[positionName] <= pos + self.delta
            ):
                self.emit("predefinedPositionChanged", (positionName, pos))
                break

    def get_predefined_positions_list(self):
        return self.predefinedPositionsNamesList

    def moveToPosition(self, positionName):
        try:
            self.move(self.predefinedPositions[positionName])
        except Exception:
            self.log.exception(
                "Cannot move motor %s: invalid position name.", str(self.username)
            )

    def get_current_position_name(self):
        if self.is_ready() and self.get_state() == self.READY:
            for positionName in self.predefinedPositions:
                if (
                    self.predefinedPositions[positionName]
                    >= self.get_value() - self.delta
                    and self.predefinedPositions[positionName]
                    <= self.get_value() + self.delta
                ):
                    return positionName

    def setNewPredefinedPosition(self, positionName, positionOffset):
        try:
            self.execute_command("setNewPosition", positionName, positionOffset)
        except AttributeError:
            self.log.exception("Cannot set new predefined position")
